import QtQuick 2.14
import QtQuick.Controls 2.14
import QtQuick.Layouts 1.14
import QtCharts 2.14

import "components"
import "theme"
import "brake-curves/calculations.js" as Calc

Item {
   // User parameters
   property real delay: delaySlider.current // seconds

   property real speed: speedSlider.current // km/h
   property real siSpeed: speed / 3.6 // m/s

   property real speedInaccuracy: speedInaccuracySlider.current // km/h
   property real siSpeedInaccuracy: 2 * speedInaccuracy / 3.6 // m/s

   property real positionInaccuracy: positionInaccuracySlider.current // meters

   property real maxBrakePower: breakPowerSlider.current // MW
   property real siMaxBrakePower: maxBrakePower * 1000000 // W

   property real trainLength: trainLengthSlider.current // m

   // calculations
   property real endConstSpeed: 0 // m
   property real startCurveOne: 0 // m

   // output
   property real relativeBreakingDistance: endConstSpeed - startCurveOne + positionInaccuracy // m
   property real absoluteBreakingDistance: endConstSpeed - trainLength // m

   property var curveOne: Calc.calculateBreakCurve(siMaxBrakePower, Calc.TOTAL_MASS, siSpeed)
   onCurveOneChanged: {
      const series = seriesOne.upperSeries;
      series.removePoints(0, series.count);

      const stepSize = Math.max(parseInt(curveOne.speeds.length / 50), 1);

      for (let i = 0; i < curveOne.speeds.length; i += stepSize) {
         series.append(curveOne.positions[i], curveOne.speeds[i] * 3.6);
      }

      startCurveOne = curveOne.positions[0];
   }

   property var curveTwo: Calc.calculateBreakCurve(siMaxBrakePower, Calc.TOTAL_MASS, siSpeed, siSpeedInaccuracy, positionInaccuracy + trainLength)
   onCurveTwoChanged: {
      const series = seriesTwo.upperSeries;
      series.removePoints(0, series.count);

      const stepSize = Math.max(parseInt(curveTwo.speeds.length / 50), 1);

      for (let i = 0; i < curveTwo.speeds.length; i += stepSize) {
         series.append(curveTwo.positions[i], curveTwo.speeds[i] * 3.6);
      }

      updateConstantSeries();
   }

   onDelayChanged: updateConstantSeries()

   function updateConstantSeries() {
      const constantSeries = seriesConstantSpeed.upperSeries;
      constantSeries.removePoints(0, constantSeries.count);

      const maxSpeed = curveTwo.speeds[0];
      const distanceTraveled = maxSpeed * delay;

      constantSeries.append(curveTwo.positions[0], maxSpeed * 3.6);
      constantSeries.append(curveTwo.positions[0] + distanceTraveled, maxSpeed * 3.6);

      positionAxis.min = 0;
      positionAxis.max = curveTwo.positions[0] + distanceTraveled;

      speedAxis.min = 0;
      speedAxis.max = curveTwo.speeds[0] * 3.6;

      endConstSpeed = curveTwo.positions[0] + distanceTraveled;
   }

   anchors {
      fill: parent
      centerIn: parent
   }

   DLRLabel {
      id: titleLabel

      scale: widgetScale
      transformOrigin: Item.Top

      anchors {
         top: parent.top
         topMargin: verticalMargin
         horizontalCenter: parent.horizontalCenter
      }

      text: "Brake Curves"
      font.pointSize: 40
      font.bold: true
   }

   Rectangle {
      id: inputs

      scale: widgetScale
      transformOrigin: Item.Top

      anchors {
         top: parent.top
         topMargin: (titleLabel.height + titleLabel.anchors.topMargin) * widgetScale
         horizontalCenter: parent.horizontalCenter
      }

      height: sliderGrid.height + verticalMargin
      width: referenceWidth - 2 * horizontalMargin

      color: DLRTheme.backgroundLightGrey
      border.color: "black"
      radius: 5

      Grid {
         id: sliderGrid
         anchors {
            top: parent.top
            left: parent.left
            right: parent.right
            topMargin: verticalMargin / 2
            leftMargin: horizontalMargin / 2
            rightMargin: horizontalMargin / 2
         }

         columns: 2
         spacing: 10

         ValueSlider {
            width: parent.width / 2 - parent.spacing

            id: delaySlider
            title: "Reaction Delay"
            unit: "s"

            tooltipText: "The worst case reaction delay the second train may need to react to a brake from the first train."

            start: 0
            current: 0
            end: 5
            step: 0.1
         }

         ValueSlider {
            width: parent.width / 2 - parent.spacing

            id: speedSlider
            title: "Initial Speed"
            unit: "km/h"

            tooltipText: "Speed of both trains before the braking procedures."

            start: 30
            current: 200
            end: 360
            step: 10
         }

         ValueSlider {
            width: parent.width / 2 - parent.spacing

            id: speedInaccuracySlider
            title: "Speed Inaccuracy"
            unit: "km/h"

            tooltipText: "Inaccuracy of the onboard speed sensors."

            start: 0
            current: 0
            end: 36
            step: 0.5
         }

         ValueSlider {
            width: parent.width / 2 - parent.spacing

            id: positionInaccuracySlider
            title: "Position Inaccuracy"
            unit: "m"

            tooltipText: "Inaccuracy of the onboard position sensors."

            start: 0
            current: 0
            end: 500
            step: 10
         }

         ValueSlider {
            width: parent.width / 2 - parent.spacing

            id: breakPowerSlider
            title: "Max Brake Power"
            unit: "MW"

            tooltipText: "The maximal brake power of both trains."

            start: 10
            current: 14
            end: 18
            step: 1
         }

         ValueSlider {
            width: parent.width / 2 - parent.spacing

            id: trainLengthSlider
            title: "Train Length"
            unit: "m"

            tooltipText: "Both trains are considered to have the same length."

            start: 0
            current: 150
            end: 300
            step: 20
         }
      }
   }

   Row {
      id: infoRow

      scale: widgetScale
      transformOrigin: Item.Top

      anchors {
         top: parent.top
         horizontalCenter: parent.horizontalCenter
         topMargin: (0.5 * verticalMargin + inputs.height) * widgetScale + inputs.anchors.topMargin
      }

      width: referenceWidth - 2 * horizontalMargin

      spacing: 2 * horizontalMargin

      Rectangle {
         width: parent.width / 2 - horizontalMargin
         height: relativeBDLabel.implicitHeight + 6

         color: DLRTheme.backgroundLightGrey
         border.color: "black"
         radius: 5

         DLRLabel {
            id: relativeBDLabel
            text: "Relative Braking Distance: " + relativeBreakingDistance.toFixed(0) + " m"
            anchors.centerIn: parent
            font.pointSize: 16
            font.bold: true
         }
      }

      Rectangle {
         width: parent.width / 2 - horizontalMargin
         height: absoluteBDLabel.implicitHeight + 6

         color: DLRTheme.backgroundLightGrey
         border.color: "black"
         radius: 5

         DLRLabel {
            id: absoluteBDLabel
            text: "Absolute Braking Distance: " + absoluteBreakingDistance.toFixed(0) + " m"
            anchors.centerIn: parent
            font.pointSize: 16
            font.bold: true
         }
      }
   }

   ChartView {
      id: curveChart

      antialiasing: true
      smooth: true

      anchors {
         top: parent.top
         horizontalCenter: parent.horizontalCenter
         topMargin: infoRow.height * widgetScale + infoRow.anchors.topMargin
         bottom: parent.bottom
         bottomMargin: infoButton.anchors.bottomMargin
      }

      width: (referenceWidth) * widgetScale - horizontalMargin

      legend {
         font {
            pointSize: 14 * widgetScale
            family: "Arial"
         }
      }

      ValueAxis {
         id: positionAxis
         titleText: "X - Position [m]"

         titleFont {
            pointSize: 14 * widgetScale
            family: "Arial"
         }

         labelsFont {
            pointSize: 12 * widgetScale
            family: "Arial"
         }

         reverse: true
         tickType: ValueAxis.TicksFixed
      }

      ValueAxis {
         id: speedAxis
         titleText: "Y - Speed [km/h]"

         titleFont {
            pointSize: 14 * widgetScale
            family: "Arial"
         }

         labelsFont {
            pointSize: 12 * widgetScale
            family: "Arial"
         }
      }

      AreaSeries {
         id: seriesConstantSpeed
         name: "Second Train Reaction Delay Phase\u00A0\u00A0\u00A0"

         upperSeries: LineSeries {}
         color: Qt.rgba(0, 1, 0, 1)
         opacity: 0.5

         axisX: positionAxis
         axisY: speedAxis
      }

      AreaSeries {
         id: seriesTwo
         name: "Second Train Braking Phase\u00A0\u00A0\u00A0"

         upperSeries: LineSeries {}
         color: Qt.rgba(1, 0, 0, 1)
         opacity: 0.5

         axisX: positionAxis
         axisY: speedAxis
      }

      AreaSeries {
         id: seriesOne
         name: "First Train Braking Phase"

         upperSeries: LineSeries {}
         color: Qt.rgba(0, 0, 1, 1)
         opacity: 0.5

         axisX: positionAxis
         axisY: speedAxis
      }
   }

   DLRButton {
      id: infoButton

      anchors {
         bottom: parent.bottom
         right: parent.right
         rightMargin: horizontalMargin
         bottomMargin: 1.5 * verticalMargin + overviewButton.height * widgetScale
      }

      scale: widgetScale
      transformOrigin: Item.BottomRight

      text: "Show Info"
      onClicked: scenarioDescription.open()
   }

   ScenarioDescription {
      id: scenarioDescription

      title: "Brake Curves"
      descriptionURL: ":/gui/brake-curves/description.html"
   }

   ReturnToOverviewButton {
      id: overviewButton
      anchors {
         bottom: parent.bottom
         right: parent.right
         rightMargin: horizontalMargin
         bottomMargin: verticalMargin
      }
   }
}