import QtQuick 2.14
import QtQuick.Controls 2.14
import QtQuick.Controls.Styles 1.4
import QtQuick.Layouts 1.14
import QtQuick.Window 2.15
import QtQuick.Extras 1.4
import QtQuick.Shapes 1.14

import "components"
import "theme"
import "coupling/simulation.js" as NGT

Item {
   id: view
   signal returnToOverview()

   property var simParams: NGT.getSimParams(couplingTime.current, headwayTime.current, fixDistance.current, switchTime.current, stopTime.current, robustness.current, gainDistribution.current / 100.0)

   Component.onCompleted: {
      restart();
   }

   DLRLabel {
      id: titleLabel

      scale: widgetScale
      transformOrigin: Item.Top

      anchors {
         top: parent.top
         topMargin: verticalMargin
         horizontalCenter: parent.horizontalCenter
      }

      text: "Coupling"
      font.pointSize: 40
      font.bold: true
      z: 1000
   }

   Rectangle {
      id: sliderGridFrame

      scale: widgetScale
      transformOrigin: Item.Top

      anchors {
         top: parent.top
         horizontalCenter: parent.horizontalCenter
         topMargin: verticalMargin / 2 + (titleLabel.height + titleLabel.anchors.topMargin) * widgetScale
      }

      height: sliderGrid.height + verticalMargin
      width: referenceWidth - (2 * horizontalMargin / widgetScale)

      color: DLRTheme.backgroundLightGrey
      border.color: "black"
      radius: 5
      z: 1000

      Grid {
         id: sliderGrid
         anchors {
            top: parent.top
            right: parent.right
            left: parent.left
            topMargin: verticalMargin / 2
            leftMargin: horizontalMargin / 2
            rightMargin: horizontalMargin / 2
         }

         height: childrenRect.height

         columns: 4
         spacing: 10
         readonly property var sliderWidth: (width - 3 * spacing) / columns

         ValueSlider {
            width: parent.sliderWidth

            id: couplingTime
            title: "Mechanical Coupling Time"
            unit: "s"

            tooltipText: "The time trains require to mechanically couple."

            start: 60
            current: 120
            end: 180
            step: 10
         }

         ValueSlider {
            width: parent.sliderWidth

            id: headwayTime
            title: "VCTS Headway"
            unit: "s"

            tooltipText: "The headway that trains inside the VCTS will keep from each other."

            start: 1
            current: 3
            end: 5
            step: 1
         }

         ValueSlider {
            width: parent.sliderWidth

            id: fixDistance
            title: "VCTS Fixed Distance"
            unit: "m"

            tooltipText: "The distance that trains inside the VCTS will keep from each other in standstill."

            start: 1
            current: 3
            end: 5
            step: 1
         }

         ValueSlider {
            width: parent.sliderWidth

            id: switchTime
            title: "Switch Time"
            unit: "s"

            tooltipText: "The time required by the switch to change position safely."

            start: 1
            current: 10
            end: 20
            step: 1
         }

         ValueSlider {
            width: parent.sliderWidth

            id: stopTime
            title: "Stop Time"
            unit: "s"

            tooltipText: "Time the trains will provide for disembark and embarking passengers."

            start: 30
            current: 60
            end: 180
            step: 10
         }

         ValueSlider {
            width: parent.sliderWidth

            id: robustness
            title: "Initial Line Robustness"
            unit: "s"

            tooltipText: "Buffer between trains, to mitgate unpredictable delays. A delay smaller than the line robustness will not impact the next train on the line."

            start: 0
            current: 300
            end: 600
            step: 10
         }

         ValueSlider {
            width: parent.sliderWidth

            id: gainDistribution
            title: "Gain Distribution"
            unit: "%"

            tooltipText: "The extra time that is gained with VCTS is distributed by increasing the line robustness or increasing the track capacity. 100% means all the gain will be used to increase track capacity."

            start: 0
            current: 50
            end: 100
            step: 10
         }
      }
   }

   Grid {
      id: resultGrid

      scale: widgetScale
      transformOrigin: Item.Top

      anchors {
         top: parent.top
         horizontalCenter: parent.horizontalCenter
         topMargin: verticalMargin / 2 + sliderGridFrame.height * widgetScale + sliderGridFrame.anchors.topMargin
      }

      width: referenceWidth - (2 * horizontalMargin / widgetScale)

      z: 1000
      columns: 4
      spacing: 0

      readonly property int rowHeight: 30

      component TableGridCell : Rectangle {
         width: parent.width / 4
         height: parent.rowHeight
         border.color: "black"
         color: DLRTheme.backgroundLightGrey
      }

      TableGridCell {
         DLRLabel {
            anchors.centerIn: parent
            text: "Results"
            font.pointSize: 16
            font.bold: true
            font.italic: true
         }
      }

      TableGridCell {
         DLRLabel {
            anchors.centerIn: parent
            text: "MCTS"
            font.pointSize: 16
            font.bold: true
         }
      }

      TableGridCell {
         DLRLabel {
            anchors.centerIn: parent
            text: "VCTS"
            font.pointSize: 16
            font.bold: true
         }
      }

      TableGridCell {
         DLRLabel {
            anchors.centerIn: parent
            text: "Change"
            font.pointSize: 16
            font.bold: true
         }
      }

      TableGridCell {
         DLRLabel {
            anchors {
               verticalCenter: parent.verticalCenter
               left: parent.left
               leftMargin: 10
            }
            text: "Robustness [mm:ss]"
            font.pointSize: 16
            font.bold: true
         }
      }

      TableGridCell {
         DLRLabel {
            anchors {
               verticalCenter: parent.verticalCenter
               right: parent.right
               rightMargin: 10
            }
            text: NGT.toMinutesTime(simParams.mctsRobustness);
            font.pointSize: 16
            width: parent.width / parent.columns;
         }
      }

      TableGridCell {
         DLRLabel {
            anchors {
               verticalCenter: parent.verticalCenter
               right: parent.right
               rightMargin: 10
            }
            text: NGT.toMinutesTime(simParams.vctsRobustness);
            font.pointSize: 16
            width: parent.width / parent.columns;
         }
      }

      TableGridCell {
         DLRLabel {
            anchors {
               verticalCenter: parent.verticalCenter
               right: parent.right
               rightMargin: 10
            }
            text: simParams.robustnessImprovement.toFixed(2) + "%";
            font.pointSize: 16
            width: parent.width / parent.columns;
         }
      }

      TableGridCell {
         DLRLabel {
            anchors {
               verticalCenter: parent.verticalCenter
               left: parent.left
               leftMargin: 10
            }
            text: "Track Density [services/hour]"
            font.pointSize: 16
            font.bold: true
         }
      }

      TableGridCell {
         DLRLabel {
            anchors {
               verticalCenter: parent.verticalCenter
               right: parent.right
               rightMargin: 10
            }
            text: simParams.mctsDensity.toFixed(2);
            font.pointSize: 16
            width: parent.width / parent.columns;
         }
      }

      TableGridCell {
         DLRLabel {
            anchors {
               verticalCenter: parent.verticalCenter
               right: parent.right
               rightMargin: 10
            }
            text: simParams.vctsDensity.toFixed(2);
            font.pointSize: 16
            width: parent.width / parent.columns;
         }
      }

      TableGridCell {
         DLRLabel {
            anchors {
               verticalCenter: parent.verticalCenter
               right: parent.right
               rightMargin: 10
            }
            text: simParams.densityImprovement.toFixed(2) + "%";
            font.pointSize: 16
            width: parent.width / parent.columns;
         }
      }
   }

   Row {
      id: timeRow

      scale: widgetScale
      transformOrigin: Item.TopRight

      anchors {
         right: parent.right
         rightMargin: horizontalMargin
         top: parent.top
         topMargin: verticalMargin / 2 + resultGrid.height * widgetScale + resultGrid.anchors.topMargin
      }

      z: 1000
      width: 0.25 * referenceWidth

      ValueSlider {
         width: parent.width - rewindButton.width

         id: simulationSpeed
         title: "Simulation Speed"
         unit: "x"
         start: 0
         current: 0
         end: 100
         step: 1
      }

      DLRButton {
         id: rewindButton
         icon.source: "../icons/rewind.png"
         onClicked: view.restart();
         width: height + 4

         anchors {
            bottom: parent.bottom
            bottomMargin: 6
         }
      }
   }

   readonly property real wagonLengthMeter: 21
   readonly property real halfWagonLengthMeter: wagonLengthMeter / 2
   readonly property real imageLengthMeters: 114
   readonly property int trackSegments: NGT.Track.end / imageLengthMeters

   readonly property int imageWidth: Window.width
   readonly property int imageHeight: Window.height

   readonly property real imageScaleW: imageWidth / 3840
   readonly property real imageScaleH: imageHeight / 2160

   Rectangle {
      anchors {
         left: parent.left
         right: parent.right
         top: parent.top
      }
      height: 2 * parent.height / 3

      color: "#739535"
   }

   Rectangle {
      anchors {
         left: parent.left
         right: parent.right
         bottom: parent.bottom
      }
      height: 1 * parent.height / 3

      color: "#517832"
   }

   Item {
      id: trackMap
      width: NGT.Track.end / imageLengthMeters * imageWidth
      height: imageHeight * 2
      transformOrigin: Item.Left

      anchors {
         verticalCenter: parent.verticalCenter
         verticalCenterOffset: parent.height / 6
      }

      function meterToPixel(meter) {
         return meter * (width / NGT.Track.end);
      }

      function meterToPercent(meter) {
         return meter / NGT.Track.end;
      }


      readonly property real leftBound: Math.min(mctsTrainA.back, vctsTrainA.back) - 500
      readonly property real rightBound: Math.max(mctsTrainB.x + mctsTrainB.width, vctsTrainB.x + vctsTrainB.width) + 500
      readonly property real visibleRegion: rightBound - leftBound

      scale: Math.min(0.7 * imageWidth / visibleRegion, 0.15)
      x: scale * -leftBound

      component TrackSection : Item {
         property real startMeters
         property real endMeters
         readonly property real lengthMeters: endMeters - startMeters
         x: startMeters / imageLengthMeters * imageWidth
         width: lengthMeters / imageLengthMeters * imageWidth
         height: parent.height
      }

      function getRandomBackground() {
         return "../images/top-down/E1-Hintergrund-TV-" + Math.floor(Math.random() * 5 + 1) + ".webp"
      }

      readonly property var ngtTrack: NGT.Track
      readonly property int extraImages: 30

      component Track : Item {
         id: trackItem
         width: trackMap.width
         height: imageHeight

         Repeater {
            model: trackSegments + trackMap.extraImages
            Image {
               x: index * imageWidth - imageWidth
               width: imageWidth
               height: imageHeight
               source: trackMap.getRandomBackground()
               asynchronous: true
               mipmap: true
            }
         }

         Image  {
            x: -imageWidth
            width: imageWidth
            height: imageHeight
            source: "../images/top-down/E2-Doppelschiene TV.webp"
            asynchronous: true
            mipmap: true
         }

         TrackSection {
            id: trackStart
            startMeters: trackMap.ngtTrack.start
            endMeters: trackMap.ngtTrack.switch - imageLengthMeters

            Repeater {
               model: trackStart.lengthMeters / imageLengthMeters
               delegate: Image  {
                  x: index * imageWidth
                  width: imageWidth
                  height: imageHeight
                  source: "../images/top-down/E2-Doppelschiene TV.webp"
                  asynchronous: true
                  mipmap: true
               }
            }
         }

         TrackSection {
            id: trackSwitch
            startMeters: trackMap.ngtTrack.switch - imageLengthMeters
            endMeters: trackMap.ngtTrack.switch
            Image {
               anchors.fill: parent
               source: "../images/top-down/E2-Weiche 2zu1-TopView.webp"
               asynchronous: true
               mipmap: true
            }
         }

         TrackSection {
            id: trackStation
            startMeters: trackMap.ngtTrack.station - 2 * imageLengthMeters
            endMeters: trackMap.ngtTrack.station

            Image {
               width: imageWidth
               height: imageHeight
               source: "../images/top-down/E6-Bahnhof-Doppel.webp"
               asynchronous: true
               mipmap: true
            }

            Image {
               x: imageWidth
               width: imageWidth
               height: imageHeight
               source: "../images/top-down/E6-Bahnhof-Doppel.webp"
               asynchronous: true
               mipmap: true
            }
         }

         TrackSection {
            id: trackEnd
            startMeters: trackMap.ngtTrack.switch
            endMeters: trackMap.ngtTrack.end

            Repeater {
               model: trackEnd.lengthMeters / imageLengthMeters
               delegate: Image  {
                  x: index * imageWidth
                  width: imageWidth
                  height: imageHeight
                  source: "../images/top-down/E2-Einzelschiene TV.webp"
                  asynchronous: true
                  mipmap: true
               }
            }
         }

         Repeater {
            model: trackMap.extraImages - 1
            Image {
               x: trackMap.width + index * imageWidth
               width: imageWidth
               height: imageHeight
               source: "../images/top-down/E2-Einzelschiene TV.webp"
               asynchronous: true
               mipmap: true
            }
         }

         readonly property real upperTrackCenter: 0.485 * imageHeight
         readonly property real lowerTrackCenter: 0.572 * imageHeight

         Path {
            id: trackPath
            startX: 0
            startY: trackItem.lowerTrackCenter

            PathLine { x: trackSwitch.x + 0.1 * imageWidth; y: trackItem.lowerTrackCenter }
            PathLine { x: trackEnd.x - 0.1 * imageWidth; y: trackItem.upperTrackCenter }
            PathLine { x: trackEnd.x + trackEnd.width; y: trackItem.upperTrackCenter }
         }

         function pointAtPercent(p) {
            return trackPath.pointAtPercent(p);
         }
      }

      component RailCarRight : Item {
         property var circleColor: "red"
         width: 712 * imageScaleW
         height: 108 * imageScaleH
         Image {
            x: -91 * imageScaleW
            y: -24 * imageScaleH
            width: sourceSize.width * imageScaleW
            height: sourceSize.height * imageScaleH
            source: "../images/top-down/E3-Triebwagen rechts.webp"
            asynchronous: true
            mipmap: true
         }

         Circle {
            readonly property var maxOpacity: 50
            readonly property var opacityRange: 40

            visible: scale > maxOpacity - opacityRange
            centerX: parent.width / 2
            centerY: parent.height / 2
            radius: 4
            color: parent.circleColor
            opacity: 1 - ((maxOpacity - scale) / opacityRange)
            scale: (1 / trackMap.scale) * widgetScale
            transformOrigin: Item.Center
         }
      }

      component Wagon : Item {
         property var circleColor: "red"
         width: 691 * imageScaleW
         height: 108 * imageScaleH
         Image {
            x: -105 * imageScaleW
            y: -24 * imageScaleH
            width: sourceSize.width * imageScaleW
            height: sourceSize.height * imageScaleH
            source: "../images/top-down/E4-Mittelwagen.webp"
            asynchronous: true
            mipmap: true
         }

         Circle {
            readonly property var maxOpacity: 50
            readonly property var opacityRange: 40

            visible: scale > maxOpacity - opacityRange
            centerX: parent.width / 2
            centerY: parent.height / 2
            radius: 4
            color: parent.circleColor
            opacity: 1 - ((maxOpacity - scale) / opacityRange)
            scale: (1 / trackMap.scale) * widgetScale
            transformOrigin: Item.Center
         }
      }

      component RailCarLeft : Item {
         property var circleColor: "red"

         width: 722 * imageScaleW
         height: 108 * imageScaleH
         Image {
            x: -83 * imageScaleW
            y: -24 * imageScaleH
            width: sourceSize.width * imageScaleW
            height: sourceSize.height * imageScaleH
            source: "../images/top-down/E5-Triebwagen links.webp"
            asynchronous: true
            mipmap: true
         }

         Circle {
            readonly property var maxOpacity: 50
            readonly property var opacityRange: 40
            visible: scale > maxOpacity - opacityRange
            centerX: parent.width / 2
            centerY: parent.height / 2
            radius: 4
            color: parent.circleColor
            opacity: 1 - ((maxOpacity - scale) / opacityRange)
            scale: (1 / trackMap.scale) * widgetScale
            transformOrigin: Item.Center
         }
      }

      Track {
         id: mctsTrack
         
         Item {
            id: mctsTrainA

            property real trackPositionMeter: 5 * wagonLengthMeter
            readonly property real front: mctsRailCarRightA.trackPosition.x
            readonly property real back: mctsRailCarLeftA.x
   
            anchors.fill: parent
            visible: back > 0
   
            RailCarRight {
               id: mctsRailCarRightA
               circleColor: DLRTheme.teaserBlue

               readonly property real trackPositionMeter: mctsTrainA.trackPositionMeter - halfWagonLengthMeter
               readonly property var trackPosition: mctsTrack.pointAtPercent(trackMap.meterToPercent(trackPositionMeter))
   
               x: trackPosition.x - width / 2
               y: trackPosition.y - height / 2
   
               readonly property var frontPoint: mctsTrack.pointAtPercent((trackPosition.x - width) / parent.width)
               readonly property var backPoint: mctsTrack.pointAtPercent(trackPosition.x / parent.width)
               readonly property real angle: Math.atan2(backPoint.y - frontPoint.y, backPoint.x - frontPoint.x)
   
               transform: Rotation {
                  origin.x: mctsRailCarRightA.width / 2
                  origin.y: mctsRailCarRightA.height / 2
                  angle: mctsRailCarRightA.angle * 180 / Math.PI
               }
            }
   
            Repeater {
               model: 2
               delegate: Wagon {
                  id: mctsRailWagon
                  circleColor: DLRTheme.teaserBlue
                  readonly property real trackPositionMeter: mctsTrainA.trackPositionMeter - wagonLengthMeter * (index + 1) - halfWagonLengthMeter
                  readonly property var trackPosition: mctsTrack.pointAtPercent(trackMap.meterToPercent(trackPositionMeter))
                  x: trackPosition.x - width / 2
                  y: trackPosition.y - height / 2

                  readonly property var frontPoint: mctsTrack.pointAtPercent((trackPosition.x - width) / parent.width)
                  readonly property var backPoint: mctsTrack.pointAtPercent(trackPosition.x / parent.width)
                  readonly property real angle: Math.atan2(backPoint.y - frontPoint.y, backPoint.x - frontPoint.x)
   
                  transform: Rotation {
                     origin.x: mctsRailWagon.width / 2
                     origin.y: mctsRailWagon.height / 2
                     angle: mctsRailWagon.angle * 180 / Math.PI
                  }
               }
            }
   
            RailCarLeft {
               id: mctsRailCarLeftA
               circleColor: DLRTheme.teaserBlue
               readonly property real trackPositionMeter: mctsTrainA.trackPositionMeter - 3 * wagonLengthMeter - halfWagonLengthMeter
               readonly property var trackPosition: mctsTrack.pointAtPercent(trackMap.meterToPercent(trackPositionMeter))
               x: trackPosition.x - width / 2
               y: trackPosition.y - height / 2
   
               readonly property var frontPoint: mctsTrack.pointAtPercent((trackPosition.x - width) / parent.width)
               readonly property var backPoint: mctsTrack.pointAtPercent(trackPosition.x / parent.width)
               readonly property real angle: Math.atan2(backPoint.y - frontPoint.y, backPoint.x - frontPoint.x)
   
               transform: Rotation {
                  origin.x: mctsRailCarLeftA.width / 2
                  origin.y: mctsRailCarLeftA.height / 2
                  angle: mctsRailCarLeftA.angle * 180 / Math.PI
               }
            }
         }

         Item {
            id: mctsTrainB
   
            property real trackPositionMeter: 84 + 100 + 84
   
            x: trackMap.meterToPixel(trackPositionMeter) - width
            y: mctsTrack.upperTrackCenter - height / 2
   
            width: childrenRect.width
            height: childrenRect.height
   
            RailCarRight {
               id: railCarRightB
               circleColor: DLRTheme.teaserBlue
               x: wagon1B.width + wagon2B.width + railCarLeftB.width
            }
   
            Wagon {
               id: wagon1B
               circleColor: DLRTheme.teaserBlue
               x: railCarLeftB.width + wagon2B.width
            }
   
            Wagon {
               id: wagon2B
               circleColor: DLRTheme.teaserBlue
               x: railCarLeftB.width
            }
   
            RailCarLeft {
               id: railCarLeftB
               circleColor: DLRTheme.teaserBlue
            }
         }
      }

      Rectangle {
         width: parent.width * 3
         height: 4 / parent.scale
         x: -parent.width
         y: imageHeight
         color: "black"
         z: 100
      }

      Track {
         id: vctsTrack
         y: imageHeight
         
         Item {
            id: vctsTrainA
   
            property real trackPositionMeter: 5 * wagonLengthMeter
            readonly property real front: vctsRailCarRightA.trackPosition.x
            readonly property real back: vctsRailCarLeftA.x
   
            anchors.fill: parent
            visible: back > 0
   
            RailCarRight {
               id: vctsRailCarRightA
               circleColor: DLRTheme.teaserYellow
   
               readonly property real trackPositionMeter: vctsTrainA.trackPositionMeter - halfWagonLengthMeter
               readonly property var trackPosition: vctsTrack.pointAtPercent(trackMap.meterToPercent(trackPositionMeter))
   
               x: trackPosition.x - width / 2
               y: trackPosition.y - height / 2
   
               readonly property var frontPoint: vctsTrack.pointAtPercent((trackPosition.x - width) / parent.width)
               readonly property var backPoint: vctsTrack.pointAtPercent(trackPosition.x / parent.width)
               readonly property real angle: Math.atan2(backPoint.y - frontPoint.y, backPoint.x - frontPoint.x)
   
               transform: Rotation {
                  origin.x: vctsRailCarRightA.width / 2
                  origin.y: vctsRailCarRightA.height / 2
                  angle: vctsRailCarRightA.angle * 180 / Math.PI
               }
            }
   
            Repeater {
               model: 2
               delegate: Wagon {
                  id: vctsRailWagon
                  circleColor: DLRTheme.teaserYellow
                  readonly property real trackPositionMeter: vctsTrainA.trackPositionMeter - wagonLengthMeter * (index + 1) - halfWagonLengthMeter
                  readonly property var trackPosition: vctsTrack.pointAtPercent(trackMap.meterToPercent(trackPositionMeter))
                  x: trackPosition.x - width / 2
                  y: trackPosition.y - height / 2

                  readonly property var frontPoint: vctsTrack.pointAtPercent((trackPosition.x - width) / parent.width)
                  readonly property var backPoint: vctsTrack.pointAtPercent(trackPosition.x / parent.width)
                  readonly property real angle: Math.atan2(backPoint.y - frontPoint.y, backPoint.x - frontPoint.x)
   
                  transform: Rotation {
                     origin.x: vctsRailWagon.width / 2
                     origin.y: vctsRailWagon.height / 2
                     angle: vctsRailWagon.angle * 180 / Math.PI
                  }
               }
            }
   
            RailCarLeft {
               id: vctsRailCarLeftA
               circleColor: DLRTheme.teaserYellow
               readonly property real trackPositionMeter: vctsTrainA.trackPositionMeter - 3 * wagonLengthMeter - halfWagonLengthMeter
               readonly property var trackPosition: vctsTrack.pointAtPercent(trackMap.meterToPercent(trackPositionMeter))
               x: trackPosition.x - width / 2
               y: trackPosition.y - height / 2
   
               readonly property var frontPoint: vctsTrack.pointAtPercent((trackPosition.x - width) / parent.width)
               readonly property var backPoint: vctsTrack.pointAtPercent(trackPosition.x / parent.width)
               readonly property real angle: Math.atan2(backPoint.y - frontPoint.y, backPoint.x - frontPoint.x)
   
               transform: Rotation {
                  origin.x: vctsRailCarLeftA.width / 2
                  origin.y: vctsRailCarLeftA.height / 2
                  angle: vctsRailCarLeftA.angle * 180 / Math.PI
               }
            }
         }
         
         Item {
            id: vctsTrainB
   
            property real trackPositionMeter: 84 + 100 + 84
   
            x: trackMap.meterToPixel(trackPositionMeter) - width
            y: vctsTrack.upperTrackCenter - height / 2
   
            width: childrenRect.width
            height: childrenRect.height
   
            RailCarRight {
               id: vctsRailCarRightB
               circleColor: DLRTheme.teaserYellow
               x: vctsWagon1B.width + vctsWagon2B.width + vctsRailCarLeftB.width
            }
   
            Wagon {
               id: vctsWagon1B
               circleColor: DLRTheme.teaserYellow
               x: vctsRailCarLeftB.width + vctsWagon2B.width
            }
   
            Wagon {
               id: vctsWagon2B
               circleColor: DLRTheme.teaserYellow
               x: vctsRailCarLeftB.width
            }
   
            RailCarLeft {
               id: vctsRailCarLeftB
               circleColor: DLRTheme.teaserYellow
            }
         }
      }
   }

   Rectangle {
      id: miniMap
      scale: widgetScale
      transformOrigin: Item.Bottom

      anchors {
         horizontalCenter: parent.horizontalCenter
         bottom: parent.bottom
         bottomMargin: verticalMargin
      }

      z: 2000

      width: 0.6 * app.referenceWidth
      height: 0.12 * app.referenceHeight
      radius: 12
      border.width: 2
      border.color: "black"
      color: DLRTheme.backgroundLightGrey

      Shape {
         id: trackShape

         anchors {
            centerIn: parent
            horizontalCenterOffset: 0.5 * horizontalMargin
         }
         width: 0.9 * parent.width
         height: 0.9 * parent.height

         function getXFromPercent(xPercent) {
            return xPercent * trackShape.width;
         }

         ShapePath {
            id: mctsLeaderPath
            startX: 0
            startY: 0.25 * trackShape.height - 3

            strokeWidth: 2
            fillColor: "transparent"
            strokeColor: "black"
            capStyle: ShapePath.RoundCap

            PathLine {
               property var xPercent: NGT.relativePosition(NGT.Track.end);
               x: trackShape.getXFromPercent(xPercent); y: mctsLeaderPath.startY
            }
         }

         ShapePath {
            id: mctsFollowerPath
            startX: 0
            startY: 0.25 * trackShape.height + 3

            strokeWidth: 2
            fillColor: "transparent"
            strokeColor: "black"
            capStyle: ShapePath.RoundCap

            PathLine {
               property var xPercent: NGT.relativePosition(NGT.Track.switch);
               x: trackShape.getXFromPercent(xPercent); y: mctsFollowerPath.startY
            }

            PathLine {
               property var xPercent: NGT.relativePosition(NGT.Track.switch + imageLengthMeters);
               x: trackShape.getXFromPercent(xPercent); y: mctsFollowerPath.startY - 6
            }

            PathLine {
               property var xPercent: NGT.relativePosition(NGT.Track.end);
               x: trackShape.getXFromPercent(xPercent); y: mctsFollowerPath.startY - 6
            }
         }

         ShapePath {
            id: vctsLeaderPath
            startX: 0
            startY: 0.75 * trackShape.height - 3

            strokeWidth: 2
            fillColor: "transparent"
            strokeColor: "black"
            capStyle: ShapePath.RoundCap

            PathLine {
               property var xPercent: NGT.relativePosition(NGT.Track.end);
               x: trackShape.getXFromPercent(xPercent); y: vctsLeaderPath.startY
            }
         }

         ShapePath {
            id: vctsFollowerPath
            startX: 0
            startY: 0.75 * trackShape.height + 3

            strokeWidth: 2
            fillColor: "transparent"
            strokeColor: "black"
            capStyle: ShapePath.RoundCap

            PathLine {
               property var xPercent: NGT.relativePosition(NGT.Track.switch);
               x: trackShape.getXFromPercent(xPercent); y: vctsFollowerPath.startY
            }

            PathLine {
               property var xPercent: NGT.relativePosition(NGT.Track.switch + imageLengthMeters);
               x: trackShape.getXFromPercent(xPercent); y: vctsFollowerPath.startY - 6
            }

            PathLine {
               property var xPercent: NGT.relativePosition(NGT.Track.end);
               x: trackShape.getXFromPercent(xPercent); y: vctsFollowerPath.startY - 6
            }
         }
      }

      Rectangle {
         anchors.centerIn: parent
         width: parent.width - 2 * parent.border.width
         height: 1
         color: "white"
      }

      Item {
         anchors.centerIn: parent
         width: parent.width
         height: 0.9 * parent.height

         DLRLabel {
            anchors {
               top: parent.top
               topMargin: 0.25 * parent.height - 0.5 * height
               left: parent.left
               leftMargin: 0.5 * horizontalMargin
            }

            text: "MCTS"
            font.pointSize: 10
            font.bold: true
         }

         Rectangle {
            id: minimapMCTSStation
            width: trackShape.getXFromPercent(NGT.relativePosition(200))
            height: 7
            border.color: "black"
            color: DLRTheme.teaserGreen

            DLRLabel {
               anchors {
                  horizontalCenter: parent.horizontalCenter
                  top: parent.bottom
               }

               text: "Station"
               font.pointSize: 8
            }
         }

         Rectangle {
            id: minimapMCTSLeader
            height: 7
            radius: 3
            width: trackShape.getXFromPercent(NGT.relativePosition(82))
            color: DLRTheme.teaserBlue
            border.color: "black"

            DLRLabel {
               anchors {
                  horizontalCenter: parent.horizontalCenter
                  bottom: parent.top
               }

               text: "1"
               font.pointSize: 8
            }
         }

         Rectangle {
            id: minimapMCTSFollower
            height: 7
            radius: 3
            width: trackShape.getXFromPercent(NGT.relativePosition(82))
            color: DLRTheme.teaserBlue
            border.color: "black"

            DLRLabel {
               anchors {
                  horizontalCenter: parent.horizontalCenter
                  bottom: parent.top
               }

               text: "2"
               font.pointSize: 8
            }
         }

         DLRLabel {
            anchors {
               top: parent.top
               topMargin: 0.75 * parent.height - 0.5 * height
               left: parent.left
               leftMargin: 0.5 * horizontalMargin
            }

            text: "VCTS"
            font.pointSize: 10
            font.bold: true
         }

         Rectangle {
            id: minimapVCTSStation
            width: trackShape.getXFromPercent(NGT.relativePosition(200))
            height: 7
            border.color: "black"
            color: DLRTheme.teaserGreen

            DLRLabel {
               anchors {
                  horizontalCenter: parent.horizontalCenter
                  top: parent.bottom
               }

               text: "Station"
               font.pointSize: 8
            }
         }

         Rectangle {
            id: minimapVCTSLeader
            height: 7
            radius: 3
            width: trackShape.getXFromPercent(NGT.relativePosition(82))
            color: DLRTheme.teaserYellow
            border.color: "black"

            DLRLabel {
               anchors {
                  horizontalCenter: parent.horizontalCenter
                  bottom: parent.top
               }

               text: "1"
               font.pointSize: 8
            }
         }

         Rectangle {
            id: minimapVCTSFollower
            height: 7
            radius: 3
            width: trackShape.getXFromPercent(NGT.relativePosition(82))
            color: DLRTheme.teaserYellow
            border.color: "black"

            DLRLabel {
               anchors {
                  horizontalCenter: parent.horizontalCenter
                  bottom: parent.top
               }

               text: "2"
               font.pointSize: 8
            }
         }
      }
   }

   function update() {
      mctsTrainA.trackPositionMeter = NGT.simulation.mctsFollower.position;
      vctsTrainA.trackPositionMeter = NGT.simulation.vctsFollower.position;

      mctsTrainB.trackPositionMeter = NGT.simulation.mctsLeader.position;
      vctsTrainB.trackPositionMeter = NGT.simulation.vctsLeader.position;

      {
         const mctsLeaderPos = NGT.relativePosition(NGT.simulation.mctsLeader.position);
         const p = mctsLeaderPath.pointAtPercent(mctsLeaderPos);
         minimapMCTSLeader.x = trackShape.x + p.x - minimapMCTSLeader.width;
         minimapMCTSLeader.y = p.y - minimapMCTSLeader.height / 2;
      }
      {
         const mctsFollowerPos = NGT.relativePosition(NGT.simulation.mctsFollower.position);
         const p = mctsFollowerPath.pointAtPercent(mctsFollowerPos);
         minimapMCTSFollower.x = trackShape.x + p.x - minimapMCTSFollower.width;
         minimapMCTSFollower.y = p.y - minimapMCTSFollower.height / 2;
      }
      {
         const vctsLeaderPos = NGT.relativePosition(NGT.simulation.vctsLeader.position);
         const p = vctsLeaderPath.pointAtPercent(vctsLeaderPos);
         minimapVCTSLeader.x = trackShape.x + p.x - minimapVCTSLeader.width;
         minimapVCTSLeader.y = p.y - minimapVCTSLeader.height / 2;
      }
      {
         const vctsFollowerPos = NGT.relativePosition(NGT.simulation.vctsFollower.position);
         const p = vctsFollowerPath.pointAtPercent(vctsFollowerPos);
         minimapVCTSFollower.x = trackShape.x + p.x - minimapVCTSFollower.width;
         minimapVCTSFollower.y = p.y - minimapVCTSFollower.height / 2;
      }

      {
         const mctsStationPos = NGT.relativePosition(NGT.Track.station);
         const p = mctsLeaderPath.pointAtPercent(mctsStationPos);
         minimapMCTSStation.x = trackShape.x + p.x - minimapMCTSStation.width;
         minimapMCTSStation.y = p.y + 2.5;
      }
      {
         const vctsStationPos = NGT.relativePosition(NGT.Track.station);
         const p = vctsLeaderPath.pointAtPercent(vctsStationPos);
         minimapVCTSStation.x = trackShape.x + p.x - minimapVCTSStation.width;
         minimapVCTSStation.y = p.y + 2.5;
      }

   }

   function restart() {
      NGT.createSimulation(couplingTime.current, headwayTime.current, fixDistance.current, switchTime.current, stopTime.current, robustness.current, gainDistribution.current / 100.0);
      simLoop.wait = 0;
      simLoop.tick = 0;
   }

   Timer {
      id: simLoop
      interval: 1
      running: true
      repeat: true

      property real wait: 0
      property var tick: 0

      onTriggered: {
         if (NGT.simulation.finished) {
            if (wait < 3000) {
               wait += interval;
            } else {
               view.restart();
            }
         } else {
            tick += 1;
            NGT.simulation.simulate((interval / 1000) * simulationSpeed.current);
            if (tick % 12 === 0) {
               update();
            }
         }
      }
   }

   DLRButton {
      anchors {
         bottom: parent.bottom
         right: parent.right
         rightMargin: horizontalMargin
         bottomMargin: 1.5 * verticalMargin + overviewButton.height * widgetScale
      }

      scale: widgetScale
      transformOrigin: Item.BottomRight

      text: "Show Info"
      onClicked: scenarioDescription.open()
   }

   ScenarioDescription {
      id: scenarioDescription
      z: 3000

      title: "Coupling"
      descriptionURL: ":/gui/coupling/description.html"

      onOpened: simulationSpeed.current = 0
      onClosed: simulationSpeed.current = 1
   }

   ReturnToOverviewButton {
      id: overviewButton
      z: 3001

      anchors {
         bottom: parent.bottom
         right: parent.right
         rightMargin: horizontalMargin
         bottomMargin: verticalMargin
      }
   }
}