/** @type m/s^2 */
const ACCELERATION = 1;

/** @type m */
const TRAIN_LENGTH = 82;

/**
 * @param speed{m/s}
 * @param position{m}
 */
function Train(speed, position) {
   /** @type m/s */
   this.speed = speed;

   /** @type m */
   this.position = position;

   return this;
}

function SimulationParameters(couplingTime, headwayTime, fixDistance, switchTime, stoppingTime, mctsRobustness, gainDistribution) {
   /** @type s */
   this.switchTime = switchTime;

   /** @type s */
   this.stoppingTime = stoppingTime;

   /** @type s */
   this.couplingTime = couplingTime;

   /** @type s */
   this.headwayTime = headwayTime;

   /** @type m */
   this.fixDistance = fixDistance;

   /** @type s */
   this.mctsRobustness = mctsRobustness;

   /** @type train/hour */
   this.mctsDensity = 3600 / (this.mctsRobustness + stoppingTime + couplingTime + switchTime + Math.sqrt((4 * TRAIN_LENGTH) / ACCELERATION));

   /** @type s */
   this.improvementTime = couplingTime - headwayTime + Math.sqrt((4 * TRAIN_LENGTH) / ACCELERATION)
      - Math.sqrt((2 * (2 * TRAIN_LENGTH + fixDistance)) / ACCELERATION);

   /** @type s */
   this.vctsRobustness = mctsRobustness + (gainDistribution * this.improvementTime);

   /** @type train/hour */
   this.vctsDensity = 3600 / ((3600 / this.mctsDensity) - (1 - gainDistribution) * this.improvementTime);

   /** @type number */
   this.densityImprovement = 100 * (this.vctsDensity - this.mctsDensity) / this.mctsDensity;

   /** @type number */
   this.robustnessImprovement = 100 * (this.vctsRobustness - mctsRobustness) / mctsRobustness;

   return this;
}

const Track = {
   /** @type m */
   start: 0,

   /** @type m */
   switch: 1482 * 2,

   /** @type m */
   station: 3990,

   /** @type m */
   end: 10032
};

/**
 * @param p{m}
 */
function relativePosition(p) {
   return Math.max(0, Math.min(1, p / Track.end));
}

/**
 * @param couplingTime{s}
 * @param headwayTime{s}
 * @param fixDistance{m}
 * @param switchTime{s}
 * @param stoppingTime{s}
 * @param mctsRobustness{s}
 * @param gainDistribution{number}
 */
function Simulation(couplingTime, headwayTime, fixDistance, switchTime, stoppingTime, mctsRobustness, gainDistribution) {
   /** @type SimulationParameters */
   this.simParameters = new SimulationParameters(couplingTime, headwayTime, fixDistance, switchTime, stoppingTime, mctsRobustness, gainDistribution);

   /** @type {m/s} */
   const initialSpeed = 50;

   /** @type Train */
   this.mctsLeader = new Train(initialSpeed, switchTime * initialSpeed + 2 * TRAIN_LENGTH);
   /** @type Train */
   this.mctsFollower = new Train(initialSpeed, TRAIN_LENGTH);

   /** @type Train */
   this.vctsLeader = new Train(initialSpeed, switchTime * initialSpeed + 2 * TRAIN_LENGTH);
   /** @type Train */
   this.vctsFollower = new Train(initialSpeed, TRAIN_LENGTH);

   /** @type boolean */
   this.finished = false;

   /** @type s */
   this.elapsedTime = 0;

   /** @param deltaTime{s} */
   this.simulate = function (deltaTime) {
      this.simulateMCTS(deltaTime);
      this.simulateVCTS(deltaTime);

      this.elapsedTime += deltaTime;

      if (vctsLeaderPhase === 5 && mctsLeaderPhase === 5) {
         this.finished = true;
      }
   }

   /**
    * @param speed{m/s}
    * @return {m}
    */
   this.breakDistance = function (speed) {
      return (speed * speed) / (2 * ACCELERATION);
   }

   const breakPoint = Track.station - this.breakDistance(initialSpeed);

   let mctsLeaderPhase = 0;
   let mctsFollowerPhase = 0;

   let mctsStopped = false;
   let mctsStopTime = 0;

   /** @param deltaTime{s} */
   this.simulateMCTS = function (deltaTime) {
      if (mctsStopped) {
         mctsStopTime += deltaTime;
      }

      switch (mctsLeaderPhase) { // simulate mcts leader
         case 0: // Drive with constant speed.
            this.mctsLeader.position += this.mctsLeader.speed * deltaTime;
            if (this.mctsLeader.position > breakPoint) {
               mctsLeaderPhase++;
            }
            break;

         case 1: // Break
            this.mctsLeader.speed -= ACCELERATION * deltaTime;
            this.mctsLeader.position += this.mctsLeader.speed * deltaTime;
            if (this.mctsLeader.speed <= 0) {
               this.mctsLeader.position = Track.station;
               this.mctsLeader.speed = 0;
               mctsLeaderPhase++;
            }
            break;

         case 2: // Wait at train station.
            if (mctsStopTime > this.simParameters.stoppingTime + this.simParameters.couplingTime) {
               mctsLeaderPhase++;
            }
            break;

         case 3: // Accelerate
            this.mctsLeader.speed += ACCELERATION * deltaTime;
            this.mctsLeader.position += this.mctsLeader.speed * deltaTime;
            if (this.mctsLeader.speed >= initialSpeed) {
               this.mctsLeader.speed = initialSpeed;
               mctsLeaderPhase++;
            }
            break;

         case 4: // Drive with constant speed.
            this.mctsLeader.position += this.mctsLeader.speed * deltaTime;
            if (this.mctsLeader.position >= Track.end) {
               this.mctsLeader.position = Track.end;
               mctsLeaderPhase++;
            }
            break;

         default: // Simulation ended for this train.
            break;

      }

      switch (mctsFollowerPhase) { // simulate mcts follower
         case 0: // Drive with constant speed.
            this.mctsFollower.position += this.mctsFollower.speed * deltaTime;
            if (this.mctsFollower.position > breakPoint - TRAIN_LENGTH) {
               mctsFollowerPhase++;
            }
            break;

         case 1: // Break
            this.mctsFollower.speed -= ACCELERATION * deltaTime;
            this.mctsFollower.position += this.mctsFollower.speed * deltaTime;
            if (this.mctsFollower.speed <= 0) {
               this.mctsFollower.position = Track.station - TRAIN_LENGTH;
               this.mctsLeader.speed = 0;
               mctsStopped = true;
               mctsFollowerPhase++;
            }
            break;

         case 2: // Wait at train station.
            if (mctsStopTime > this.simParameters.stoppingTime + this.simParameters.couplingTime) {
               mctsFollowerPhase++;
            }
            break;

         case 3: // Accelerate
            this.mctsFollower.speed += ACCELERATION * deltaTime;
            this.mctsFollower.position += this.mctsFollower.speed * deltaTime;
            if (this.mctsFollower.speed >= initialSpeed) {
               this.mctsFollower.speed = initialSpeed;
               mctsFollowerPhase++;
            }
            break;

         case 4: // Drive with constant speed.
            this.mctsFollower.position += this.mctsFollower.speed * deltaTime;
            if (this.mctsFollower.position >= Track.end - TRAIN_LENGTH) {
               this.mctsFollower.position = Track.end - TRAIN_LENGTH;
               mctsFollowerPhase++;
            }
            break;

         default: // Simulation ended for this train.
            break;

      }

   }

   let vctsLeaderPhase = 0;
   let vctsFollowerPhase = 0;

   let vctsStopped = false;
   let vctsStopTime = 0;

   /** @param deltaTime{s} */
   this.simulateVCTS = function (deltaTime) {
      if (vctsStopped) {
         vctsStopTime += deltaTime;
      }

      switch (vctsLeaderPhase) { // simulate mcts leader
         case 0: // Drive with constant speed.
            this.vctsLeader.position += this.vctsLeader.speed * deltaTime;
            if (this.vctsLeader.position > breakPoint) {
               vctsLeaderPhase++;
            }
            break;

         case 1: // Break
            this.vctsLeader.speed -= ACCELERATION * deltaTime;
            this.vctsLeader.position += this.vctsLeader.speed * deltaTime;
            if (this.vctsLeader.speed <= 0) {
               this.vctsLeader.position = Track.station;
               this.vctsLeader.speed = 0;
               vctsLeaderPhase++;
            }
            break;

         case 2: // Wait at train station.
            if (vctsStopTime > this.simParameters.stoppingTime) {
               vctsLeaderPhase++;
            }
            break;

         case 3: // Accelerate
            this.vctsLeader.speed += ACCELERATION * deltaTime;
            this.vctsLeader.position += this.vctsLeader.speed * deltaTime;
            if (this.vctsLeader.speed >= initialSpeed) {
               this.vctsLeader.speed = initialSpeed;
               vctsLeaderPhase++;
            }
            break;

         case 4: // Drive with constant speed.
            this.vctsLeader.position += this.vctsLeader.speed * deltaTime;
            if (this.vctsLeader.position >= Track.end) {
               this.vctsLeader.position = Track.end;
               vctsLeaderPhase++;
            }
            break;

         default: // Simulation ended for this train.
            break;

      }

      switch (vctsFollowerPhase) { // simulate vcts follower
         case 0: // Drive with constant speed.
            this.vctsFollower.position += this.vctsFollower.speed * deltaTime;
            if (this.vctsFollower.position > breakPoint - TRAIN_LENGTH - this.simParameters.fixDistance) {
               vctsFollowerPhase++;
            }
            break;

         case 1: // Break
            this.vctsFollower.speed -= ACCELERATION * deltaTime;
            this.vctsFollower.position += this.vctsFollower.speed * deltaTime;
            if (this.vctsFollower.speed <= 0) {
               this.vctsFollower.position = Track.station - TRAIN_LENGTH - this.simParameters.fixDistance;
               this.vctsFollower.speed = 0;
               vctsStopped = true;
               vctsFollowerPhase++;
            }
            break;

         case 2: // Wait at train station.
            if (vctsStopTime > this.simParameters.stoppingTime + this.simParameters.headwayTime) {
               vctsFollowerPhase++;
            }
            break;

         case 3: // Accelerate
            this.vctsFollower.speed += ACCELERATION * deltaTime;
            this.vctsFollower.position += this.vctsFollower.speed * deltaTime;
            if (this.vctsFollower.speed >= initialSpeed) {
               this.vctsFollower.speed = initialSpeed;
               vctsFollowerPhase++;
            }
            break;

         case 4: // Drive with constant speed.
            this.vctsFollower.position += this.vctsFollower.speed * deltaTime;
            if (this.vctsFollower.position >= Track.end - TRAIN_LENGTH) {
               this.vctsFollower.position = Track.end - TRAIN_LENGTH;
               vctsFollowerPhase++;
            }
            break;

         default: // Simulation ended for this train.
            break;

      }
   }
}

let simulation = new Simulation(120, 3, 3, 5, 60, 300, 0.5);

/**
 * @param couplingTime{s}
 * @param headwayTime{s}
 * @param fixDistance{m}
 * @param switchTime{s}
 * @param stoppingTime{s}
 * @param mctsRobustness{s}
 * @param gainDistribution{number}
 */
function createSimulation(couplingTime, headwayTime, fixDistance, switchTime, stoppingTime, mctsRobustness, gainDistribution) {
   simulation = new Simulation(couplingTime, headwayTime, fixDistance, switchTime, stoppingTime, mctsRobustness, gainDistribution);
}

function getSimParams(couplingTime, headwayTime, fixDistance, switchTime, stoppingTime, mctsRobustness, gainDistribution) {
   return new SimulationParameters(couplingTime, headwayTime, fixDistance, switchTime, stoppingTime, mctsRobustness, gainDistribution);
}

function toMinutesTime(time) {
   time = Math.floor(time);
   const minutes = Math.floor(time / 60).toString();
   const seconds = (time % 60).toString();
   return `${minutes.padStart(2, '0')}:${seconds.padStart(2, '0')}`;
}