import {infoTexts} from "./dsc-info-texts.mjs"

/** @type m/s^2 */
const ACCELERATION = 1;

/** @type m */
const SHORT_TRAIN_LENGTH = 84;

/** @type m */
const LONG_TRAIN_LENGTH = 252;

/** @type m/s */
const TARGET_SPEED = 200 / 3.6;

/** @type m/s */
const LONG_TRAIN_SPEED_AFTER_STATION = 120 / 3.6;

/** @type s */
const STOPPING_TIME = 10;

/** @type s */
const SWITCH_TIME = 5;

/** @type m */
const HEADWAY = 100

/** @type s */
const HEADWAY_TIME = HEADWAY / TARGET_SPEED;

class Train {
   /**
    * @param speed{m/s}
    * @param position{m}
    */
   constructor(speed, position) {
      /** @type m/s */
      this.speed = speed;

      /** @type m */
      this.position = position;
   }
}

export const Track = {
   /** @type m */
   start: 0,

   /** @type m */
   entrySwitch: 2052,

   /** @type m */
   station: 4218,

   /** @type m */
   exitSwitch: 4446,

   /** @type m */
   end: 10032
};

/**
 * @param d{m}
 * @return {m}
 */
export function relativePosition(d) {
   return d / Track.end;
}

class Simulation {
   constructor() {
      /** @type m */
      const separation = TARGET_SPEED * HEADWAY_TIME;

      /** @type Train */
      this.trainA = new Train(TARGET_SPEED, LONG_TRAIN_LENGTH);
      /** @type Train */
      this.trainB = new Train(TARGET_SPEED, this.trainA.position + separation + SHORT_TRAIN_LENGTH);
      /** @type Train */
      this.trainC = new Train(TARGET_SPEED, this.trainB.position + separation + SHORT_TRAIN_LENGTH);
      /** @type Train */
      this.trainD = new Train(TARGET_SPEED, this.trainC.position + separation + SHORT_TRAIN_LENGTH);

      /** @type boolean */
      this.finished = false;

      /** @type s */
      this.elapsedTime = 0;

      this.trainAPhase = 0;
      this.trainBPhase = 0;
      this.trainCPhase = 0;
      this.trainDPhase = 0;

      /** @type s */
      this.entrySwitchSwitchTime = 0;

      /** @type boolean */
      this.entrySwitchStartedSwitching = false;

      /** @type s */
      this.exitSwitchSwitchTime = 0;

      this.stopTime = 0;

      this.infoText = "";
      this.infoTextIndex = 0;

      this.statusText = "";
      this.statusTextIndex = 0;
   }

   /** @param deltaTime{s} */
   simulate(deltaTime) {
      this.elapsedTime += deltaTime;

      if (this.infoTextIndex < infoTexts.length && infoTexts[this.infoTextIndex].t < this.elapsedTime) {
         this.infoText = infoTexts[this.infoTextIndex++].text;
      }

      this.simulateShortTrains(deltaTime);
      this.simulateLongTrain(deltaTime);
   }

   /** @param deltaTime{s} */
   simulateShortTrains(deltaTime) {
      switch (this.trainBPhase) {
         case 0: // Drive with constant speed.
            this.trainB.position += this.trainB.speed * deltaTime;

            if (this.trainB.position - SHORT_TRAIN_LENGTH > Track.entrySwitch) {
               this.entrySwitchStartedSwitching = true;
            }

            if (Track.station - this.trainB.position - 2 * SHORT_TRAIN_LENGTH < breakDistance(this.trainB.speed)) {
               this.trainBPhase++;
            }
            break;

         case 1: // Break
            this.trainB.speed -= ACCELERATION * deltaTime;
            this.trainB.position += this.trainB.speed * deltaTime;
            if (this.trainB.speed <= 0) {
               this.trainB.speed = 0;
               this.trainBPhase++;
            }
            break;

         case 2:
            if (this.exitSwitchSwitchTime >= SWITCH_TIME + 2 * HEADWAY_TIME) {
               this.trainBPhase++;
            }
            break;

         case 3:
            this.trainB.speed += ACCELERATION * deltaTime;
            this.trainB.position += this.trainB.speed * deltaTime;
            if (this.trainB.speed >= TARGET_SPEED) {
               this.trainB.speed = TARGET_SPEED;
               this.trainBPhase++;
            }
            break;

         case 4:
            this.trainB.position += this.trainB.speed * deltaTime;
            break;
      }

      if (this.entrySwitchStartedSwitching) {
         this.entrySwitchSwitchTime += deltaTime;
      }

      switch (this.trainCPhase) {
         case 0: // Drive with constant speed.
            this.trainC.position += this.trainC.speed * deltaTime;

            if (Track.station - this.trainC.position - SHORT_TRAIN_LENGTH < breakDistance(this.trainC.speed)) {
               this.trainCPhase++;
            }
            break;

         case 1: // Break
            this.trainC.speed -= ACCELERATION * deltaTime;
            this.trainC.position += this.trainC.speed * deltaTime;
            if (this.trainC.speed <= 0) {
               this.trainC.speed = 0;
               this.trainCPhase++;
            }
            break;

         case 2:
            if (this.exitSwitchSwitchTime >= SWITCH_TIME + HEADWAY_TIME) {
               this.trainCPhase++;
            }
            break;

         case 3:
            this.trainC.speed += ACCELERATION * deltaTime;
            this.trainC.position += this.trainC.speed * deltaTime;
            if (this.trainC.speed >= TARGET_SPEED) {
               this.trainC.speed = TARGET_SPEED;
               this.trainCPhase++;
            }
            break;

         case 4:
            this.trainC.position += this.trainC.speed * deltaTime;
            break;

      }

      switch (this.trainDPhase) {
         case 0: // Drive with constant speed.
            this.trainD.position += this.trainD.speed * deltaTime;

            if (Track.station - this.trainD.position < breakDistance(this.trainD.speed)) {
               this.trainDPhase++;
            }
            break;

         case 1: // Break
            this.trainD.speed -= ACCELERATION * deltaTime;
            this.trainD.position += this.trainD.speed * deltaTime;
            if (this.trainD.speed <= 0) {
               this.trainD.speed = 0;
               this.trainDPhase++;
            }
            break;

         case 2:
            if (this.exitSwitchSwitchTime >= SWITCH_TIME) {
               this.trainDPhase++;
            }
            break;

         case 3:
            this.trainD.speed += ACCELERATION * deltaTime;
            this.trainD.position += this.trainD.speed * deltaTime;
            if (this.trainD.speed >= TARGET_SPEED) {
               this.trainD.speed = TARGET_SPEED;
               this.trainDPhase++;
            }
            break;

         case 4:
            this.trainD.position += this.trainD.speed * deltaTime;
            break;
      }
   }

   /** @param deltaTime{s} */
   simulateLongTrain(deltaTime) {
      switch (this.trainAPhase) {
         case 0: // Drive with constant speed.
            this.trainA.position += this.trainA.speed * deltaTime;
            if (Track.entrySwitch - this.trainA.position < breakDistance(this.trainA.speed)) {
               this.trainAPhase++;
            }
            break;

         case 1: // Break.
            this.trainA.speed = Math.max(this.trainA.speed - ACCELERATION * deltaTime, 0);
            this.trainA.position += this.trainA.speed * deltaTime;
            if (this.entrySwitchSwitchTime >= SWITCH_TIME) {
               this.trainAPhase++;
            }
            break;

         case 2: // Accelerate again
            this.trainA.speed += ACCELERATION * deltaTime;
            this.trainA.position += this.trainA.speed * deltaTime;

            // Target speed reached
            if (this.trainA.speed >= TARGET_SPEED) {
               this.trainA.speed = TARGET_SPEED;
               this.trainAPhase++;
            }

            // Need to break for station
            if (Track.station - this.trainA.position < breakDistance(this.trainA.speed)) {
               this.trainAPhase += 2;
            }
            break;

         case 3: // Drive with constant speed.
            this.trainA.position += this.trainA.speed * deltaTime;
            if (Track.station - this.trainA.position < breakDistance(this.trainA.speed)) {
               this.trainAPhase++;
            }
            break;

         case 4: // Break.
            this.trainA.speed -= ACCELERATION * deltaTime;
            this.trainA.position += this.trainA.speed * deltaTime;
            if (this.trainA.speed <= 0) {
               this.trainA.speed = 0;
               this.trainAPhase++;
            }
            break;

         case 5: // Passenger Exchange
            this.stopTime += deltaTime;
            if (this.stopTime > STOPPING_TIME) {
               this.trainAPhase++;
            }
            break;

         case 6: // Accelerate again
            this.trainA.speed += ACCELERATION * deltaTime;
            this.trainA.position += this.trainA.speed * deltaTime;

            if (this.trainA.speed >= LONG_TRAIN_SPEED_AFTER_STATION) {
               this.trainAPhase++;
            }
            break;

         case 7: // Drive with constant speed
            this.trainA.position += this.trainA.speed * deltaTime;

            const trainABack = this.trainA.position - LONG_TRAIN_LENGTH;
            if (trainABack >= Track.exitSwitch) {
               this.exitSwitchSwitchTime += deltaTime;
            }

            if (trainABack - this.trainD.position - HEADWAY < breakDistance(this.trainD.speed - this.trainA.speed)
               && this.trainA.speed < this.trainD.speed) {
               this.trainAPhase++;
            }

            break;

         case 8:
            this.trainA.speed += ACCELERATION * deltaTime;
            this.trainA.position += this.trainA.speed * deltaTime;
            this.exitSwitchSwitchTime += deltaTime;

            if (this.trainA.speed >= TARGET_SPEED) {
               this.trainA.speed = TARGET_SPEED;
               this.trainAPhase++;
            }
            break;

         case 9:
            this.trainA.position += this.trainA.speed * deltaTime;

            if (this.trainA.position >= Track.end) {
               this.trainA.position = Track.end;
               this.trainAPhase++;
            }
            break;

         default:
            this.finished = true;
            return;

      }
   }
}

/**
 * @param speed{m/s}
 * @return {m}
 */
function breakDistance(speed) {
   return (speed * speed) / (2 * ACCELERATION);
}

export let simulation = new Simulation();

export function createSimulation() {
   simulation = new Simulation();
}

export function toMinutesTime(time) {
   time = Math.floor(time);
   const minutes = Math.floor(time / 60).toString();
   const seconds = (time % 60).toString();
   return `${minutes.padStart(2, '0')}:${seconds.padStart(2, '0')}`;
}