import {infoTexts} from "./ssc-info-texts.mjs";

/** @type m/s^2 */
const ACCELERATION = 1;

/** @type m */
const TRAIN_LENGTH = 100;

/** @type m/s */
const TARGET_SPEED = 200 / 3.6;

/** @type s */
const STOPPING_TIME = 10;

/** @type s */
const HEADWAY_TIME = 1;

class Train {
   /**
    * @param speed{m/s}
    * @param position{m}
    */
   constructor(speed, position) {
      /** @type m/s */
      this.speed = speed;

      /** @type m */
      this.position = position;
   }
}

export const Track = {
   /** @type m */
   start: 0,

   /** @type m */
   station: 2000,

   /** @type m */
   end: 4500
};

/**
 * @param d{m}
 * @return {{x: m, y: m}}
 */
export function relativePosition(d) {
   return d / Track.end;
}

class Simulation {
   constructor() {
      /** @type Train */
      this.leader = new Train(TARGET_SPEED, TARGET_SPEED * HEADWAY_TIME + TRAIN_LENGTH);
      /** @type Train */
      this.main = new Train(TARGET_SPEED, 0);

      const mainStopGoTime = STOPPING_TIME + (2 * TARGET_SPEED);
      const followerStopGoTime = ((2 * (TARGET_SPEED * TARGET_SPEED)) / (2 * ACCELERATION)) / TARGET_SPEED;
      const stopGoDiff = mainStopGoTime - followerStopGoTime;

      /** @type Train */
      this.follower = new Train(TARGET_SPEED, -(TARGET_SPEED * (HEADWAY_TIME + stopGoDiff) + TRAIN_LENGTH));

      /** @type boolean */
      this.finished = false;

      /** @type s */
      this.elapsedTime = 0;

      this.breakPoint = Track.station - breakDistance(TARGET_SPEED);

      this.phase = 0;
      this.stopTime = 0;

      this.infoText = "";
      this.infoTextIndex = 0;

      this.statusText = "";
      this.statusTextIndex = 0;
   }

   /** @param deltaTime{s} */
   simulate(deltaTime) {
      this.elapsedTime += deltaTime;

      if (this.infoTextIndex < infoTexts.length && infoTexts[this.infoTextIndex].t < this.elapsedTime) {
         this.infoText = infoTexts[this.infoTextIndex++].text;
      }

      switch (this.phase) {
         case 0: // Drive with constant speed.
            this.main.position += this.main.speed * deltaTime;
            if (this.main.position > this.breakPoint) {
               this.phase++;
            }
            break;

         case 1: // Break
            this.main.speed -= ACCELERATION * deltaTime;
            this.main.position += this.main.speed * deltaTime;
            if (this.main.speed <= 0) {
               // this.main.position = Track.station;
               this.main.speed = 0;
               this.phase++;
            }
            break;

         case 2: // Wait at train station.
            this.stopTime += deltaTime;
            if (this.stopTime > STOPPING_TIME) {
               this.phase++;
            }
            break;

         case 3: // Accelerate
            this.main.speed += ACCELERATION * deltaTime;
            this.main.position += this.main.speed * deltaTime;
            if (this.main.speed >= TARGET_SPEED) {
               this.main.speed = TARGET_SPEED;
               this.phase++;
            }
            break;

         case 4: // Drive with constant speed.
            this.main.position += this.main.speed * deltaTime;
            if (this.main.position >= Track.end) {
               this.main.position = Track.end;
               this.phase++;
            }
            break;

         default: // Simulation ended for this train.
            this.finished = true;
            return;
      }

      this.leader.position += this.leader.speed * deltaTime;
      this.follower.position += this.follower.speed * deltaTime;
   }
}

/**
 * @param speed{m/s}
 * @return {m}
 */
function breakDistance(speed) {
   return (speed * speed) / (2 * ACCELERATION);
}

export let simulation = new Simulation();

export function createSimulation() {
   simulation = new Simulation();
}

export function toMinutesTime(time) {
   time = Math.floor(time);
   const minutes = Math.floor(time / 60).toString();
   const seconds = (time % 60).toString();
   return `${minutes.padStart(2, '0')}:${seconds.padStart(2, '0')}`;
}