import QtQuick 2.14
import QtQuick.Controls 2.14
import QtQuick.Controls 1.4
import QtQuick.Controls.Styles 1.4
import QtQuick.Layouts 1.14
import QtQuick.Window 2.15
import QtQuick.Extras 1.4
import QtQuick.Shapes 1.14

import "../components"
import "../theme"
import "simulation-ssc.mjs" as NGT

Item {
   id: view
   anchors.fill: parent

   Component.onCompleted: NGT.createSimulation();

   readonly property int imageWidth: Window.width
   readonly property int imageHeight: Window.height
   readonly property real imageScale: imageHeight / 1080
   readonly property real imageLengthMeters: 39

   function pixelsForDistance(dist) {
      return Math.round(dist * (hstMain.width / 50));
   }

   readonly property real panStart: 500
   property real pan: -panStart

   function render(elapsed) {
      foreground.update();
      background.update();

      hstLead.anchors.horizontalCenterOffset = (pixelsForDistance(NGT.simulation.leader.position - NGT.simulation.main.position) + view.pan) / 3;
      hstFollow.anchors.horizontalCenterOffset = (pixelsForDistance(NGT.simulation.follower.position - NGT.simulation.main.position) + view.pan) / 3;
      hstMain.anchors.horizontalCenterOffset = view.pan;

      station.anchors.horizontalCenterOffset = 2 * pixelsForDistance(NGT.Track.station - NGT.simulation.main.position) + view.pan;

      if (NGT.simulation.elapsedTime > 10 && view.pan < 0) {
         view.pan += 10 * elapsed;
      }

      if (NGT.simulation.elapsedTime > 75 && view.pan < panStart) {
         view.pan += 10 * elapsed;
      }

      infoText.text = NGT.simulation.infoText;
      // timeDisplay.text = NGT.toMinutesTime(NGT.simulation.elapsedTime);
      speedDisplay.value = (NGT.simulation.main.speed * 3.6);

      // minimap
      {
         const minimapLeaderPos = NGT.relativePosition(NGT.simulation.leader.position);
         if (0 <= minimapLeaderPos && minimapLeaderPos <= 1) {
            const p = trainPath.pointAtPercent(minimapLeaderPos);
            minimapLeader.x = trackShape.x + p.x - minimapLeader.width;
            minimapLeader.y = p.y - minimapLeader.height / 2;
            minimapLeader.visible = true;
         } else {
            minimapLeader.visible = false;
         }
      }
      {
         const minimapMainPos = NGT.relativePosition(NGT.simulation.main.position);
         const p = trainPath.pointAtPercent(minimapMainPos);
         minimapMain.x = trackShape.x + p.x - minimapMain.width;
         minimapMain.y = p.y - minimapMain.height / 2;
      }
      {
         const minimapFollowerPos = NGT.relativePosition(NGT.simulation.follower.position);
         if (0 <= minimapFollowerPos && minimapFollowerPos <= 1) {
            const p = trainPath.pointAtPercent(minimapFollowerPos);
            minimapFollower.x = trackShape.x + p.x - minimapFollower.width;
            minimapFollower.y = p.y - minimapFollower.height / 2;
            minimapFollower.visible = true;
         } else {
            minimapFollower.visible = false;
         }
      }
      {
         const stationPos = NGT.relativePosition(NGT.Track.station);
         const p = trainPath.pointAtPercent(stationPos);
         minimapStation.x = trackShape.x + p.x - minimapStation.width;
         minimapStation.y = p.y - minimapStation.height - 2.5
      }
   }

   Row {
      id: timeRow
      scale: widgetScale
      transformOrigin: Item.TopRight

      anchors {
         top: parent.top
         right: parent.right
         rightMargin: horizontalMargin
         topMargin: verticalMargin + (titleLabel.height * widgetScale) + titleLabel.anchors.topMargin
      }

      z: 100
      width: 0.25 * referenceWidth

      ValueSlider {
         id: speedSlider
         width: parent.width - rewindButton.width

         title: "Simulation Speed"
         unit: "x"

         start: 0
         current: 0
         end: 10
         step: 1
      }

      DLRButton {
         id: rewindButton
         icon.source: "../../icons/rewind.png"
         onClicked: view.restart();
         width: height + 4

         anchors {
            bottom: parent.bottom
            bottomMargin: 6
         }
      }
   }

   CircularGauge {
      Rectangle {
         width: parent.width + 2
         height: parent.height + 2
         radius: width / 2
         color: "black"
         border.color: "white"
         border.width: 2
         z: -1
      }

      id: speedDisplay

      anchors {
         bottom: miniMap.top
         bottomMargin: verticalMargin * widgetScale
         horizontalCenter: parent.horizontalCenter
      }

      scale: widgetScale * 0.4
      transformOrigin: Item.Bottom
      z: 100

      minimumValue: 0
      maximumValue: 300

      style: CircularGaugeStyle {
         tickmarkStepSize: 30
      }
   }

   Rectangle {
      scale: widgetScale
      transformOrigin: Item.Center
      z: 100
      anchors {
         centerIn: parent
         verticalCenterOffset: -0.15 * imageHeight
      }
      width: 500
      height: infoText.height + 20
      radius: 10
      border.width: 2
      border.color: "black"
      color: DLRTheme.backgroundLightGrey

      DLRLabel {
         id: infoText
         anchors {
            centerIn: parent
         }
         width: parent.width - 20
         horizontalAlignment: Text.AlignHCenter
         verticalAlignment: Text.AlignVCenter

         wrapMode: Text.Wrap
         font.pointSize: 14
      }
   }

   Item {
      id: background
      width: (NGT.Track.end / imageLengthMeters * imageWidth) / 3
      height: imageHeight

      x: 0
      anchors.bottom: parent.bottom

      function getRandomImage() {
         return "../../images/parallax/Background-" + Math.floor(Math.random() * 5 + 1) + ".webp"
      }

      function update() {
         x = (-pixelsForDistance(NGT.simulation.main.position) + view.pan) / 3;
      }

      Repeater {
         model: NGT.Track.end / imageLengthMeters
         Image {
            x: index * imageWidth
            width: imageWidth
            height: imageHeight
            source: background.getRandomImage()
            asynchronous: true
            mipmap: true
         }
      }
   }

   Item {
      id: foreground
      width: NGT.Track.end / imageLengthMeters * imageWidth
      height: imageHeight
      anchors.bottom: parent.bottom

      function getRandomImage() {
         return "../../images/parallax/Foreground-" + Math.floor(Math.random() * 4 + 1) + ".webp"
      }

      x: 0

      function update() {
         x =  2 * -pixelsForDistance(NGT.simulation.main.position) + view.pan;
      }

      Repeater {
         model: NGT.Track.end / imageLengthMeters
         Image {
            x: index * imageWidth
            width: imageWidth
            height: imageHeight
            source: foreground.getRandomImage()
            asynchronous: true
            mipmap: true
         }
      }
   }

   Image {
      id: station
      anchors {
         bottom: parent.bottom;
         bottomMargin: 0.29 * imageHeight;
         horizontalCenter: parent.horizontalCenter;
      }
      source: "../../images/Station-large.png";
      width: 1500 * imageScale;
      height: 125 * imageScale;
   }

   Image {
      id: hstMain
      anchors {
         bottom: parent.bottom;
         bottomMargin: 0.265 * imageHeight;
         horizontalCenter: parent.horizontalCenter;
      }
      source: "../../images/HST_Complete.png";
      width: 1200 * imageScale;
      height: 105 * imageScale;

      Text {
         anchors {
            centerIn: parent
            verticalCenterOffset: -10 * imageScale
            horizontalCenterOffset: -parent.width / 3.5
         }
         text: "Unit 2"
         font {
            pointSize: 15 * imageScale
            family: "Bahnschrift SemiBold"
            bold: true
            italic: true
         }
      }
      Text {
         anchors {
            centerIn: parent
            verticalCenterOffset: -10 * imageScale
            horizontalCenterOffset: parent.width / 3.5
         }
         text: "Unit 2"
         font {
            pointSize: 15 * imageScale
            family: "Bahnschrift SemiBold"
            bold: true
            italic: true
         }
      }
   }

   Image {
      id: hstFollow
      anchors {
         bottom: parent.bottom;
         bottomMargin: 0.265 * imageHeight;
         horizontalCenter: parent.horizontalCenter;
      }
      source: "../../images/HST_Complete.png";
      width: 1200 * imageScale;
      height: 105 * imageScale;


      Text {
         anchors {
            centerIn: parent
            verticalCenterOffset: -10 * imageScale
            horizontalCenterOffset: parent.width / 3.5
         }
         text: "Unit 3"
         font {
            pointSize: 15 * imageScale
            family: "Bahnschrift SemiBold"
            bold: true
            italic: true
         }
      }
   }

   Image {
      id: hstLead
      anchors {
         bottom: parent.bottom;
         bottomMargin: 0.265 * imageHeight;
         horizontalCenter: parent.horizontalCenter;
      }
      source: "../../images/HST_Complete.png";
      width: 1200 * imageScale;
      height: 105 * imageScale;

      Text {
         anchors {
            centerIn: parent
            verticalCenterOffset: -10 * imageScale
            horizontalCenterOffset: -parent.width / 3.5
         }
         text: "Unit 1"
         font {
            pointSize: 15 * imageScale
            family: "Bahnschrift SemiBold"
            bold: true
            italic: true
         }
      }
   }

   Rectangle {
      id: miniMap
      scale: widgetScale
      transformOrigin: Item.Bottom

      anchors {
         horizontalCenter: parent.horizontalCenter
         bottom: parent.bottom
         bottomMargin: verticalMargin
      }

      width: 0.6 * app.referenceWidth
      height: 0.06 * app.referenceHeight
      radius: 12
      border.width: 2
      border.color: "black"
      color: DLRTheme.backgroundLightGrey

      Shape {
         id: trackShape

         anchors.centerIn: parent
         width: 0.9 * parent.width
         height: 0.9 * parent.height

         function getXFromPercent(xPercent) {
            return xPercent * trackShape.width;
         }

         ShapePath {
            id: trainPath
            startX: 0
            startY: 0.5 * trackShape.height

            strokeWidth: 2
            fillColor: "transparent"
            strokeColor: "black"
            capStyle: ShapePath.RoundCap

            PathLine {
               property var xPercent: NGT.relativePosition(NGT.Track.end);
               x: trackShape.getXFromPercent(xPercent); y: trainPath.startY
            }
         }
      }

      Item {
         anchors.centerIn: parent
         width: parent.width
         height: 0.9 * parent.height

         Rectangle {
            id: minimapLeader
            height: 7
            radius: 3
            width: trackShape.getXFromPercent(NGT.relativePosition(100))
            color: DLRTheme.teaserYellow
            border.color: "black"

            DLRLabel {
               anchors {
                  horizontalCenter: parent.horizontalCenter
                  top: parent.bottom
               }

               text: "U1"
               font.pointSize: 8
            }
         }

         Rectangle {
            id: minimapMain
            height: 7
            radius: 3
            width: trackShape.getXFromPercent(NGT.relativePosition(100))
            color: DLRTheme.teaserYellow
            border.color: "black"

            DLRLabel {
               anchors {
                  horizontalCenter: parent.horizontalCenter
                  top: parent.bottom
               }

               text: "U2"
               font.pointSize: 8
            }
         }

         Rectangle {
            id: minimapFollower
            height: 7
            radius: 3
            width: trackShape.getXFromPercent(NGT.relativePosition(100))
            color: DLRTheme.teaserYellow
            border.color: "black"

            DLRLabel {
               anchors {
                  horizontalCenter: parent.horizontalCenter
                  top: parent.bottom
               }

               text: "U3"
               font.pointSize: 8
            }
         }

         Rectangle {
            id: minimapStation
            width: trackShape.getXFromPercent(NGT.relativePosition(100))
            height: 7
            border.color: "black"
            color: DLRTheme.teaserGreen

            DLRLabel {
               anchors {
                  horizontalCenter: parent.horizontalCenter
                  bottom: parent.top
               }

               text: "Station"
               font.pointSize: 8
            }
         }
      }
   }

   function restart() {
      NGT.createSimulation();
      simLoop.wait = 0;
      pan = -panStart;
   }

   Timer {
      id: simLoop
      interval: 12
      running: true
      repeat: true

      property real wait: 0

      onTriggered: {
         if (NGT.simulation.finished) {
            if (wait < 3000) {
               wait += interval;
            } else {
               NGT.createSimulation();
               wait = 0;
               pan = -panStart;
            }
         } else {
            const elapsed = (interval / 1000) * speedSlider.current;
            NGT.simulation.simulate(elapsed);
            render(elapsed);
         }
      }
   }

   DLRButton {
      anchors {
         bottom: parent.bottom
         right: parent.right
         rightMargin: horizontalMargin
         bottomMargin: 1.5 * verticalMargin + overviewButton.height * widgetScale
      }

      scale: widgetScale
      transformOrigin: Item.BottomRight

      text: "Show Info"
      onClicked: scenarioDescription.open()
   }

   ScenarioDescription {
      id: scenarioDescription
      z: 1000

      title: "Single Slip Coaching"
      descriptionURL: ":/gui/slip-coaching/ssc-description.html"

      onOpened: speedSlider.current = 0
      onClosed: speedSlider.current = 1
   }

   ReturnToOverviewButton {
      id: overviewButton
      z: 1001
      anchors {
         bottom: parent.bottom
         right: parent.right
         rightMargin: horizontalMargin
         bottomMargin: verticalMargin
      }
   }
}