import QtQuick 2.14
import QtQuick.Controls 2.14
import QtQuick.Layouts 1.14
import QtQuick.Shapes 1.14
import QtQuick.Window 2.15

import SpeedLimits 1.0
import "speed-limits"
import "speed-limits/components"
import "components"
import "theme"

Item {
   readonly property int imageWidth: Window.width
   readonly property int imageHeight: Window.height

   property int trackStart: 0.05 * width
   property int trackEnd: 0.95 * width

   property int vctsTrackY: height - (height / 10)

   property real leftBound: speedLimitsModel.bounds.x
   property real rightBound: speedLimitsModel.bounds.y
   property real visibleBound: Math.max(rightBound - leftBound, 1)

   function pixelPosition(positionKm) {
      const positionInBoundsKm = positionKm - leftBound;
      const relativePosition = positionInBoundsKm / visibleBound;

      const trackPixelLength = trackEnd - trackStart;
      return relativePosition * trackPixelLength + trackStart;
   }

   function pixelLength(lengthKm) {
      const relativePosition = lengthKm / visibleBound;
      const trackPixelLength = trackEnd - trackStart;
      return relativePosition * trackPixelLength;
   }

   function pixelHeight(heightMeter) {
      const trackMeterLength = visibleBound * 1000;
      const trackPixelLength = trackEnd - trackStart;
      const pixelPerMeter = trackPixelLength / trackMeterLength;

      return heightMeter * pixelPerMeter;
   }

   anchors {
      fill: parent
      centerIn: parent
   }

   DLRLabel {
      id: titleLabel

      scale: widgetScale
      transformOrigin: Item.Top

      anchors {
         top: parent.top
         topMargin: verticalMargin
         horizontalCenter: parent.horizontalCenter
      }

      text: "Speed Limits"
      font.pointSize: 40
      font.bold: true
      z: 1000
   }

   id: view
   signal returnToOverview()

   SpeedLimitsModel {
      id: speedLimitsModel
      property real trackLength: track.totalLength * 1000 + 1800

      function relativePosition(p) {
         return Math.max(0, Math.min(1, p / trackLength));
      }
   }

   Timer {
      interval: 8
      running: true
      repeat: true

      onTriggered: {
         speedLimitsModel.simulate(interval);
         update();
      }
   }

   function update() {
      {
         const stationPos = speedLimitsModel.relativePosition(speedLimitsModel.trackLength);
         const p = mctsPath.pointAtPercent(stationPos);
         minimapStation.x = trackShape.x + p.x - minimapStation.width;
      }

      {
         const mctsPos = speedLimitsModel.relativePosition(speedLimitsModel.mcts.position * 1000);
         const p = mctsPath.pointAtPercent(mctsPos);
         minimapMCTS.x = trackShape.x + p.x - minimapMCTS.width;
      }

      for (let i = 0; i < minimapVCTS.count; i++) {
         const vcts = minimapVCTS.itemAt(i);
         const trainPos = speedLimitsModel.relativePosition(vcts.train.position * 1000);
         const p = vctsPath.pointAtPercent(trainPos);
         vcts.x = trackShape.x + p.x - vcts.width;
         vcts.visible = trainPos > 0;
      }
   }

   Rectangle {
      width: parent.width
      anchors {
         top: parent.top
         bottom: background.top
      }

      color: "#8AC4F2"
   }

   Image {
      id: background
      source: "../images/parallax/Background-1.webp"
      asynchronous: true
      anchors {
         bottom: parent.bottom
         bottomMargin: -height / 2.9
      }
      width: parent.width
      fillMode: Image.PreserveAspectFit
      mipmap: true
   }

   Rectangle {
      width: parent.width
      anchors {
         top: foreground.bottom
         bottom: parent.bottom
      }

      gradient: Gradient {
        GradientStop { position: 0.0; color: "#4A7E4D" }
        GradientStop { position: 1.0; color: "#3D683F" }
      }
   }

   Foreground {
      id: foreground
      anchors {
         bottom: parent.bottom
         bottomMargin: parent.height / 6
      }
   }

   ListView {
      id: speedLimitSigns

      anchors.bottom: foreground.bottom

      model: speedLimitsModel.track
      delegate: TrackSegment {
         start: pixelPosition(model.start)
         length: pixelLength(model.length)
         speedLimit: model.speedLimit
      }
   }

   property real trainSegmentLength: 50
   property real trainSegmentHeight: 12.5

   Train {
      id: mctsTrain
      labelPos: "top"
      circleColor: DLRTheme.teaserBlue

      anchors {
         bottom: foreground.bottom
         bottomMargin: 0.310 * foreground.height
      }

      numberOfTrains: speedLimitsModel.vcts.numTrains
      numberOfSegments: (speedLimitsModel.mcts.length / numberOfTrains) / trainSegmentLength

      start: pixelPosition(speedLimitsModel.mcts.position)
      length: pixelLength(speedLimitsModel.mcts.length / 1000.0)
      speed: speedLimitsModel.mcts.speed
      pHeight: pixelHeight(trainSegmentHeight)
   }

   Image {
      width: pixelLength(6 * 0.3)
      height: pixelHeight(20)
      x: pixelPosition(65)
      source: "../images/Station.png"
      anchors {
         bottom: foreground.bottom
         bottomMargin: 0.292 * foreground.height
      }
   }

   ListView {
      id: vctsList
      model: speedLimitsModel.vcts

      anchors {
         bottom: foreground.bottom
         bottomMargin: 0.273 * foreground.height
      }

      delegate: Train {
         labelPos: "bottom"
         circleColor: DLRTheme.teaserYellow
         numberOfTrains: 1
         numberOfSegments: model.length / trainSegmentLength

         start: pixelPosition(model.position)
         length: pixelLength(model.length / 1000.0)
         speed: model.speed
         pHeight: pixelHeight(trainSegmentHeight)
      }
   }


   Rectangle {
      id: miniMap
      scale: widgetScale
      transformOrigin: Item.Bottom

      anchors {
         horizontalCenter: parent.horizontalCenter
         bottom: parent.bottom
         bottomMargin: verticalMargin
      }

      z: 2000

      width: 0.6 * app.referenceWidth
      height: 0.06 * app.referenceHeight
      radius: 12
      border.width: 2
      border.color: "black"
      color: DLRTheme.backgroundLightGrey

      Shape {
         id: trackShape

         anchors.centerIn: parent
         width: 0.925 * parent.width
         height: 0.9 * parent.height

         function getXFromPercent(xPercent) {
            return xPercent * trackShape.width;
         }

         ShapePath {
            id: mctsPath
            startX: 0
            startY: 0.5 * trackShape.height - 7

            strokeWidth: 2
            fillColor: "transparent"
            strokeColor: "black"
            capStyle: ShapePath.RoundCap

            PathLine {
               x: trackShape.width; y: mctsPath.startY
            }
         }

         ShapePath {
            id: vctsPath
            startX: 0
            startY: 0.5 * trackShape.height + 7

            strokeWidth: 2
            fillColor: "transparent"
            strokeColor: "black"
            capStyle: ShapePath.RoundCap

            PathLine {
               x: trackShape.width; y: vctsPath.startY
            }
         }
      }

      Item {
         anchors.centerIn: parent
         width: parent.width
         height: 0.9 * parent.height

         Repeater {
            id: mmSpeedLimits
            anchors.fill: parent

            model: speedLimitsModel.track
            delegate: Item {
               property var limit: model
               y: 0.5 * mmSpeedLimits.height - height / 2

               x: trackShape.x + trackShape.getXFromPercent(speedLimitsModel.relativePosition(model.start * 1000))
               width: trackShape.getXFromPercent(speedLimitsModel.relativePosition(model.length * 1000))

               Rectangle {
                  anchors {
                     verticalCenter: parent.verticalCenter
                     left: parent.left
                  }
                  width: 1
                  height: 25

                  color: "red"
               }

               DLRLabel {
                  id: mmSpeedLimitLabel

                  anchors.centerIn: parent
                  text: model.speedLimit
                  font.pointSize: 7
               }

               Rectangle {
                  anchors {
                     verticalCenter: parent.verticalCenter
                     right: parent.right
                  }
                  width: 1
                  height: 25

                  color: "red"
               }
            }
         }

         Rectangle {
            id: minimapStation
            anchors.verticalCenter: parent.verticalCenter

            width: trackShape.getXFromPercent(speedLimitsModel.relativePosition(1800))
            height: 7
            border.color: "black"
            color: DLRTheme.teaserGreen

            DLRLabel {
               anchors {
                  verticalCenter: parent.verticalCenter
                  left: parent.right
                  leftMargin: 2
               }

               text: "Station"
               font.pointSize: 7
            }
         }

         Rectangle {
            id: minimapMCTS
            height: 7
            radius: 3
            y: (0.5 * trackShape.height - 7) - height / 2;
            width: Math.max(trackShape.getXFromPercent(speedLimitsModel.relativePosition(speedLimitsModel.mcts.length)), 6)
            color: DLRTheme.teaserBlue
            border.color: "black"

            DLRLabel {
               anchors {
                  horizontalCenter: parent.horizontalCenter
                  bottom: parent.top
               }

               text: "MCTS"
               font.pointSize: 7
            }
         }

         Repeater {
            id: minimapVCTS
            anchors.fill: parent
            model: speedLimitsModel.vcts
            delegate: Rectangle {
               property var train: model
               height: 7
               y: (0.5 * trackShape.height + 7) - height / 2;
               radius: 3
               width: Math.max(trackShape.getXFromPercent(speedLimitsModel.relativePosition(model.length)), 6)
               color: DLRTheme.teaserYellow
               border.color: "black"

               DLRLabel {
                  anchors {
                     horizontalCenter: parent.horizontalCenter
                     top: parent.bottom
                  }

                  text: index + 1
                  font.pointSize: 7
               }
            }
         }
      }
   }

   NextSpeedLimits {
      id: nextTrackList
      scale: widgetScale
      transformOrigin: Item.TopLeft

      anchors {
         top: parent.top
         left: parent.left
         leftMargin: horizontalMargin
         topMargin: verticalMargin + (titleLabel.height + titleLabel.anchors.topMargin) * widgetScale
      }

      height: childrenRect.height
      width: 0.5 * referenceWidth
   }

   TimeControl {
      id: timeCol
      scale: widgetScale
      transformOrigin: Item.TopRight

      anchors {
         top: parent.top
         right: parent.right
         rightMargin: horizontalMargin
         topMargin: verticalMargin + (titleLabel.height + titleLabel.anchors.topMargin) * widgetScale
      }

      width: 0.25 * referenceWidth
   }

   TrainProperties {
      id: trainProps
      scale: widgetScale
      transformOrigin: Item.TopRight

      anchors {
         top: parent.top
         right: parent.right
         rightMargin: horizontalMargin
         topMargin: verticalMargin + timeCol.height * widgetScale + timeCol.anchors.topMargin
      }

      width: 0.25 * referenceWidth
   }

   Column {
      scale: widgetScale
      transformOrigin: Item.TopRight

      anchors {
         top: parent.top
         right: parent.right
         rightMargin: horizontalMargin
         topMargin: verticalMargin + trainProps.height * widgetScale + trainProps.anchors.topMargin
      }

      width: 0.25 * referenceWidth
      spacing: 15

      DLRLabel {
         text: "Time to the station (mm:ss):"
         font.pointSize: 16
         font.bold: true
      }

      DLRLabel {
         property int minutes: parseInt(speedLimitsModel.mctsTime / 60)
         property int seconds: speedLimitsModel.mctsTime - (minutes * 60)
         text: "MCTS:\t" + (minutes + ":").padStart(3, '0') + (seconds + "").padStart(2, '0')
         font.pointSize: 16
      }

      DLRLabel {
         property int minutes: parseInt(speedLimitsModel.vctsTime / 60)
         property int seconds: speedLimitsModel.vctsTime - (minutes * 60)
         text: "VCTS:\t" + (minutes + ":").padStart(3, '0') + (seconds + "").padStart(2, '0')
         font.pointSize: 16
      }
   }

   DLRButton {
      anchors {
         bottom: parent.bottom
         right: parent.right
         rightMargin: horizontalMargin
         bottomMargin: 1.5 * verticalMargin + overviewButton.height * widgetScale
      }

      scale: widgetScale
      transformOrigin: Item.BottomRight

      text: "Show Info"
      onClicked: scenarioDescription.open()
   }

   ScenarioDescription {
      id: scenarioDescription
      z: 3000

      title: "Speed Limits"
      descriptionURL: ":/gui/speed-limits/description.html"

      onOpened: timeCol.setSpeed(0)
      onClosed: timeCol.setSpeed(10)
   }

   ReturnToOverviewButton {
      id: overviewButton
      z: 3001

      anchors {
         right: parent.right
         bottom: parent.bottom
         rightMargin: horizontalMargin
         bottomMargin: verticalMargin
      }
   }
}