import QtQuick 2.14
import QtQuick.Controls 2.14
import QtQuick.Controls.Styles 1.4
import QtQuick.Layouts 1.14
import QtQuick.Window 2.15
import QtQuick.Extras 1.4
import QtQuick.Shapes 1.14

import "components"
import "theme"
import "switches"
import "switches/simulation.js" as NGT

Item {
   id: view
   anchors.fill: parent

   DLRLabel {
      id: titleLabel

      scale: widgetScale
      transformOrigin: Item.Top

      anchors {
         top: parent.top
         topMargin: verticalMargin
         horizontalCenter: parent.horizontalCenter
      }

      text: "Switches"
      font.pointSize: 40
      font.bold: true
      z: 1000
   }

   ButtonGroup {
      id: radioGroup
      exclusive: true
   }

   Column {
      scale: widgetScale
      transformOrigin: Item.TopLeft
      width: 250
      spacing: 5
      z: 1000

      anchors {
         top: parent.top
         left: parent.left
         leftMargin: horizontalMargin
         topMargin: (verticalMargin + titleLabel.height + titleLabel.anchors.topMargin) * widgetScale
      }

      component SwitchButton : Row {
         property string text
         property string tooltipText
         property bool checked: false

         width: parent.width
         height: buttonLabel.implicitHeight
         spacing: 5

         DLRButton {
            id: buttonLabel
            text: parent.text
            checked: parent.checked
            checkable: true
            ButtonGroup.group: radioGroup
            width: parent.width - infoPopup.width - parent.spacing
         }

         InfoPopup {
            id: infoPopup
            tooltipText: parent.tooltipText
         }
      }

      SwitchButton {
         checked: true
         text: "Traditional"
         tooltipText: "Traditional Switches. They require some time to change position and during the change the track section gets blocked."
      }

      SwitchButton {
         text: "Smart"
         tooltipText: "The switches can communicate with the train directly. Therefore trains know the exact time when the switch will be released and can approach the switch according to its own braking capability. (The train must always be sure that in case the switch fail to change, it will be able to brake before the switch)."
      }

      SwitchButton {
         text: "Passive"
         tooltipText: "The merging switch does not require any switching time and is safe by design: it does not need to confirm that it managed to change positions. The separating switch works like a smart switch."
      }

      SwitchButton {
         text: "Failsafe"
         tooltipText: "Both switches have some switching time, but they assume that they always succeed in changing the position: trains can target its front end to pass over the switch directly after the switch changed its position."
      }
   }

   Column {
      scale: widgetScale
      transformOrigin: Item.TopRight
      anchors {
         top: parent.top
         right: parent.right
         topMargin: (verticalMargin + titleLabel.height + titleLabel.anchors.topMargin) * widgetScale
         rightMargin: horizontalMargin
      }

      width: 0.25 * referenceWidth
      z: 1000

      spacing: 5

      ValueSlider {
         width: parent.width - parent.spacing

         id: speed
         title: "Train Initial Speed"
         unit: "km/h"

         tooltipText: "Initial speed that the trains will approach the station."

         start: 120
         current: 200
         end: 320
         step: 10
      }

      ValueSlider {
         width: parent.width - parent.spacing

         id: switchTime
         title: "Switch Time"
         unit: "s"

         tooltipText: "The time the switch requires to change its position. In the Passive mode, the switch time only applies for the first (separating) switch."

         start: 5
         current: 10
         end: 20
         step: 1
      }

      DLRLabel {
         id: frequencyText

         anchors.horizontalCenter: parent.horizontalCenter

         property real frequency: NGT.getFrequency(NGT.Scenario[radioGroup.checkedButton.text.toUpperCase()], speed.current, switchTime.current) * 60 * 60
         text: "Maximal Track Density: " + frequency.toFixed(0) + " services/hour"
         font.pointSize: 16

         topPadding: 20
         bottomPadding: 20
      }

      Row {
         id: timeRow

         z: 1000
         width: parent.width - parent.spacing

         ValueSlider {
            width: parent.width - parent.spacing - rewindButton.width

            id: simulationSpeed
            title: "Simulation Speed"
            unit: "x"
            start: 0
            current: 0
            end: 100
            step: 1
         }

         DLRButton {
            id: rewindButton
            icon.source: "../icons/rewind.png"
            onClicked: view.restart();
            width: height + 4

            anchors {
               bottom: parent.bottom
               bottomMargin: 6
            }
         }
      }
   }

   readonly property real wagonLengthMeter: 21
   readonly property real halfWagonLengthMeter: wagonLengthMeter / 2
   readonly property real imageLengthMeters: 114

   readonly property int imageWidth: Window.width
   readonly property int imageHeight: Window.height

   readonly property real imageScaleW: imageWidth / 3840
   readonly property real imageScaleH: imageHeight / 2160

   Rectangle {
      anchors {
         left: parent.left
         right: parent.right
         top: parent.top
      }
      height: 2 * parent.height / 3

      color: "#739535"
   }

   Rectangle {
      anchors {
         left: parent.left
         right: parent.right
         bottom: parent.bottom
      }
      height: 1 * parent.height / 3

      color: "#517832"
   }

   Item {
      id: trackMap
      width: NGT.Track.end / imageLengthMeters * imageWidth
      height: imageHeight
      transformOrigin: Item.Left

      anchors {
         verticalCenter: parent.verticalCenter
         verticalCenterOffset: parent.height / 6
      }

      function meterToPixel(meter) {
         return meter * (width / NGT.Track.end);
      }

      function meterToPercent(meter) {
         return meter / NGT.Track.end;
      }

      readonly property real leftBound: Math.min(trackTrainA.back, trackTrainB.x) - 500
      readonly property real rightBound: Math.max(trackTrainA.front, trackTrainB.x + trackTrainB.width) + 500
      readonly property real visibleRegion: rightBound - leftBound

      scale: imageWidth / visibleRegion
      x: scale * -leftBound

      component TrackSection : Item {
         property real startMeters
         property real endMeters
         readonly property real lengthMeters: endMeters - startMeters
         x: startMeters / imageLengthMeters * imageWidth
         width: lengthMeters / imageLengthMeters * imageWidth
         height: parent.height
      }

      function getRandomBackground() {
         return "../images/top-down/E1-Hintergrund-TV-" + Math.floor(Math.random() * 5 + 1) + ".webp"
      }

      Repeater {
         model: NGT.Track.end / imageLengthMeters
         Image {
            x: index * imageWidth
            width: imageWidth
            height: imageHeight
            source: trackMap.getRandomBackground()
            asynchronous: true
            mipmap: true
         }
      }

      TrackSection {
         id: trackStart
         startMeters: NGT.Track.start
         endMeters: NGT.Track.entrySwitch

         Repeater {
            model: trackStart.lengthMeters / imageLengthMeters
            delegate: Image  {
               x: index * imageWidth
               width: imageWidth
               height: imageHeight
               source: "../images/top-down/E2-Einzelschiene TV.webp"
               asynchronous: true
               mipmap: true
            }
         }
      }

      TrackSection {
         id: trackEntrySwitch
         startMeters: NGT.Track.entrySwitch
         endMeters: NGT.Track.entrySwitch + imageLengthMeters
         Image {
            anchors.fill: parent
            source: "../images/top-down/E2-Weiche 1zu2-TopView.webp"
            asynchronous: true
            mipmap: true
         }
      }

      TrackSection {
         id: trackBeforeStation
         startMeters: NGT.Track.entrySwitch + imageLengthMeters
         endMeters: NGT.Track.station - imageLengthMeters

         Repeater {
            model: trackBeforeStation.lengthMeters / imageLengthMeters
            delegate: Image  {
               x: index * imageWidth
               width: imageWidth
               height: imageHeight
               source: "../images/top-down/E2-Doppelschiene TV.webp"
               asynchronous: true
               mipmap: true
            }
         }
      }

      TrackSection {
         id: trackStation
         startMeters: NGT.Track.station - imageLengthMeters
         endMeters: NGT.Track.station

         Item {
            width: imageWidth
            height: imageHeight
            Image {
               anchors.fill: parent
               source: "../images/top-down/E2-Doppelschiene TV.webp"
               asynchronous: true
               mipmap: true
            }
            Image {
               anchors.fill: parent
               source: "../images/top-down/E6-Bahnhof-Doppel.webp"
               asynchronous: true
               mipmap: true
            }
         }
      }

      TrackSection {
         id: trackAfterStation
         startMeters: NGT.Track.station
         endMeters: NGT.Track.exitSwitch - imageLengthMeters
         Repeater {
            model: trackAfterStation.lengthMeters / imageLengthMeters
            delegate: Image  {
               x: index * imageWidth
               width: imageWidth
               height: imageHeight
               source: "../images/top-down/E2-Doppelschiene TV.webp"
               asynchronous: true
               mipmap: true
            }
         }
      }

      TrackSection {
         id: trackExitSwitch
         startMeters: NGT.Track.exitSwitch - imageLengthMeters
         endMeters: NGT.Track.exitSwitch
         Image {
            anchors.fill: parent
            source: "../images/top-down/E2-Weiche 2zu1-TopView.webp"
            asynchronous: true
            mipmap: true
         }
      }

      TrackSection {
         id: trackEnd
         startMeters: NGT.Track.exitSwitch
         endMeters: NGT.Track.end

         Repeater {
            model: trackEnd.lengthMeters / imageLengthMeters
            delegate: Image  {
               x: index * imageWidth
               width: imageWidth
               height: imageHeight
               source: "../images/top-down/E2-Einzelschiene TV.webp"
               asynchronous: true
               mipmap: true
            }
         }
      }

      Signal {
         id: entrySignal
         scale: widgetScale / parent.scale

         x: trackEntrySwitch.x
         y: trackMap.upperTrackCenter - 0.2 * height * scale
         transformOrigin: Item.Bottom
      }

      Signal {
         id: exitSignal
         scale: widgetScale / parent.scale

         x: trackEnd.x
         y: trackMap.upperTrackCenter - 0.2 * height * scale
         transformOrigin: Item.Bottom
      }

      readonly property real upperTrackCenter: 0.485 * imageHeight
      readonly property real lowerTrackCenter: 0.572 * imageHeight

      Path {
         id: trackPath
         startX: 0
         startY: trackMap.upperTrackCenter

         PathLine { x: trackEntrySwitch.x + 0.1 * imageWidth; y: trackMap.upperTrackCenter }
         PathLine { x: trackBeforeStation.x - 0.1 * imageWidth; y: trackMap.lowerTrackCenter }
         PathLine { x: trackExitSwitch.x + 0.1 * imageWidth; y: trackMap.lowerTrackCenter }
         PathLine { x: trackEnd.x - 0.1 * imageWidth; y: trackMap.upperTrackCenter }
         PathLine { x: trackEnd.x + trackEnd.width; y: trackMap.upperTrackCenter }
      }

      component RailCarRight : Item {
         property var circleColor: "red"

         width: 712 * imageScaleW
         height: 108 * imageScaleH
         Image {
            x: -91 * imageScaleW
            y: -24 * imageScaleH
            width: sourceSize.width * imageScaleW
            height: sourceSize.height * imageScaleH
            source: "../images/top-down/E3-Triebwagen rechts.webp"
            asynchronous: true
            mipmap: true
         }

         Circle {
            readonly property var maxOpacity: 50
            readonly property var opacityRange: 40
            visible: scale > maxOpacity - opacityRange
            centerX: parent.width / 2
            centerY: parent.height / 2
            radius: 4
            color: parent.circleColor
            opacity: 1 - ((maxOpacity - scale) / opacityRange)
            scale: (1 / trackMap.scale) * widgetScale
            transformOrigin: Item.Center
         }
      }

      component Wagon : Item {
         property var circleColor: "red"

         width: 691 * imageScaleW
         height: 108 * imageScaleH
         Image {
            x: -105 * imageScaleW
            y: -24 * imageScaleH
            width: sourceSize.width * imageScaleW
            height: sourceSize.height * imageScaleH
            source: "../images/top-down/E4-Mittelwagen.webp"
            asynchronous: true
            mipmap: true
         }

         Circle {
            readonly property var maxOpacity: 50
            readonly property var opacityRange: 40
            visible: scale > maxOpacity - opacityRange
            centerX: parent.width / 2
            centerY: parent.height / 2
            radius: 4
            color: parent.circleColor
            opacity: 1 - ((maxOpacity - scale) / opacityRange)
            scale: (1 / trackMap.scale) * widgetScale
            transformOrigin: Item.Center
         }
      }

      component RailCarLeft : Item {
         property var circleColor: "red"

         width: 722 * imageScaleW
         height: 108 * imageScaleH
         Image {
            x: -83 * imageScaleW
            y: -24 * imageScaleH
            width: sourceSize.width * imageScaleW
            height: sourceSize.height * imageScaleH
            source: "../images/top-down/E5-Triebwagen links.webp"
            asynchronous: true
            mipmap: true
         }

         Circle {
            readonly property var maxOpacity: 50
            readonly property var opacityRange: 40
            visible: scale > maxOpacity - opacityRange
            centerX: parent.width / 2
            centerY: parent.height / 2
            radius: 4
            color: parent.circleColor
            opacity: 1 - ((maxOpacity - scale) / opacityRange)
            scale: (1 / trackMap.scale) * widgetScale
            transformOrigin: Item.Center
         }
      }

      Item {
         id: trackTrainA

         property real trackPositionMeter: 5 * wagonLengthMeter
         readonly property real front: railCarRightA.trackPosition.x
         readonly property real back: railCarLeftA.x

         anchors.fill: parent

         RailCarRight {
            id: railCarRightA
            circleColor: DLRTheme.teaserYellow

            readonly property real trackPositionMeter: trackTrainA.trackPositionMeter - halfWagonLengthMeter
            readonly property var trackPosition: trackPath.pointAtPercent(trackMap.meterToPercent(trackPositionMeter))

            x: trackPosition.x - width / 2
            y: trackPosition.y - height / 2

            readonly property var frontPoint: trackPath.pointAtPercent((trackPosition.x - width) / parent.width)
            readonly property var backPoint: trackPath.pointAtPercent(trackPosition.x / parent.width)
            readonly property real angle: Math.atan2(backPoint.y - frontPoint.y, backPoint.x - frontPoint.x)

            transform: Rotation {
               origin.x: railCarRightA.width / 2
               origin.y: railCarRightA.height / 2
               angle: railCarRightA.angle * 180 / Math.PI
            }
         }

         Repeater {
            model: 2
            delegate: Wagon {
               id: railWagon
               circleColor: DLRTheme.teaserYellow
               readonly property real trackPositionMeter: trackTrainA.trackPositionMeter - wagonLengthMeter * (index + 1) - halfWagonLengthMeter
               readonly property var trackPosition: trackPath.pointAtPercent(trackMap.meterToPercent(trackPositionMeter))
               x: trackPosition.x - width / 2
               y: trackPosition.y - height / 2

               readonly property var frontPoint: trackPath.pointAtPercent((trackPosition.x - width) / parent.width)
               readonly property var backPoint: trackPath.pointAtPercent(trackPosition.x / parent.width)
               readonly property real angle: Math.atan2(backPoint.y - frontPoint.y, backPoint.x - frontPoint.x)

               transform: Rotation {
                  origin.x: railWagon.width / 2
                  origin.y: railWagon.height / 2
                  angle: railWagon.angle * 180 / Math.PI
               }
            }
         }

         RailCarLeft {
            id: railCarLeftA
            circleColor: DLRTheme.teaserYellow
            readonly property real trackPositionMeter: trackTrainA.trackPositionMeter - 3 * wagonLengthMeter - halfWagonLengthMeter
            readonly property var trackPosition: trackPath.pointAtPercent(trackMap.meterToPercent(trackPositionMeter))
            x: trackPosition.x - width / 2
            y: trackPosition.y - height / 2

            readonly property var frontPoint: trackPath.pointAtPercent((trackPosition.x - width) / parent.width)
            readonly property var backPoint: trackPath.pointAtPercent(trackPosition.x / parent.width)
            readonly property real angle: Math.atan2(backPoint.y - frontPoint.y, backPoint.x - frontPoint.x)

            transform: Rotation {
               origin.x: railCarLeftA.width / 2
               origin.y: railCarLeftA.height / 2
               angle: railCarLeftA.angle * 180 / Math.PI
            }
         }
      }

      Repeater {
         id: followerBrakePoints
         model: 20

         property real brakeDistanceMeter: 0

         delegate: Circle {
            readonly property real trackPositionMeter: trackTrainA.trackPositionMeter + (index / 19) * followerBrakePoints.brakeDistanceMeter
            readonly property var trackPosition: trackPath.pointAtPercent(trackMap.meterToPercent(trackPositionMeter))

            transformOrigin: Item.Center
            scale: widgetScale

            centerX: trackPosition.x
            centerY: trackPosition.y
            radius: (4 / (index / 16 + 1)) / trackMap.scale
            color: "red"
         }
      }

      Item {
         id: trackTrainB

         property real trackPositionMeter: 84 + 100 + 84

         x: trackMap.meterToPixel(trackPositionMeter) - width
         y: trackMap.upperTrackCenter - height / 2

         width: childrenRect.width
         height: childrenRect.height

         RailCarRight {
            id: railCarRightB
            circleColor: DLRTheme.teaserYellow
            x: wagon1B.width + wagon2B.width + railCarLeftB.width
         }

         Wagon {
            id: wagon1B
            circleColor: DLRTheme.teaserYellow
            x: railCarLeftB.width + wagon2B.width
         }

         Wagon {
            id: wagon2B
            circleColor: DLRTheme.teaserYellow
            x: railCarLeftB.width
         }

         RailCarLeft {
            id: railCarLeftB
            circleColor: DLRTheme.teaserYellow
         }
      }

      Repeater {
         id: leaderBrakePoints
         model: 20

         property real brakeDistanceMeter: 0
         readonly property real brakeDistancePixel: trackMap.meterToPixel(brakeDistanceMeter)

         delegate: Circle {
            transformOrigin: Item.Center
            scale: widgetScale

            centerX: trackTrainB.x + trackTrainB.width + (index / 19) * leaderBrakePoints.brakeDistancePixel
            centerY: trackMap.upperTrackCenter
            radius: (4 / (index / 16 + 1)) / trackMap.scale
            color: "red"
         }
      }
   }


   Rectangle {
      id: miniMap
      scale: widgetScale
      transformOrigin: Item.Bottom

      anchors {
         horizontalCenter: parent.horizontalCenter
         bottom: parent.bottom
         bottomMargin: verticalMargin
      }

      z: 2000

      width: 0.6 * app.referenceWidth
      height: 0.06 * app.referenceHeight
      radius: 12
      border.width: 2
      border.color: "black"
      color: DLRTheme.backgroundLightGrey

      Shape {
         id: trackShape

         anchors.centerIn: parent
         width: 0.9 * parent.width
         height: 0.9 * parent.height

         function getXFromPercent(xPercent) {
            return xPercent * trackShape.width;
         }

         ShapePath {
            id: leaderPath
            startX: 0
            startY: 0.5 * trackShape.height - 6

            strokeWidth: 2
            fillColor: "transparent"
            strokeColor: "black"
            capStyle: ShapePath.RoundCap

            PathLine {
               property var xPercent: NGT.relativePosition(NGT.Track.end);
               x: trackShape.getXFromPercent(xPercent); y: leaderPath.startY
            }
         }

         ShapePath {
            id: followerPath
            startX: 0
            startY: 0.5 * trackShape.height - 6

            strokeWidth: 2
            fillColor: "transparent"
            strokeColor: "black"
            capStyle: ShapePath.RoundCap

            PathLine {
               property var xPercent: NGT.relativePosition(NGT.Track.entrySwitch);
               x: trackShape.getXFromPercent(xPercent); y: followerPath.startY
            }

            PathLine {
               property var xPercent: NGT.relativePosition(NGT.Track.entrySwitch + imageLengthMeters);
               x: trackShape.getXFromPercent(xPercent); y: followerPath.startY + 12
            }

            PathLine {
               property var xPercent: NGT.relativePosition(NGT.Track.exitSwitch - imageLengthMeters);
               x: trackShape.getXFromPercent(xPercent); y: followerPath.startY + 12
            }

            PathLine {
               property var xPercent: NGT.relativePosition(NGT.Track.exitSwitch);
               x: trackShape.getXFromPercent(xPercent); y: followerPath.startY
            }

            PathLine {
               property var xPercent: NGT.relativePosition(NGT.Track.end);
               x: trackShape.getXFromPercent(xPercent); y: followerPath.startY
            }
         }
      }

      Item {
         anchors.centerIn: parent
         width: parent.width
         height: 0.9 * parent.height

         Rectangle {
            id: minimapStation
            width: 18
            height: 6
            border.color: "black"
            color: DLRTheme.teaserGreen

            Component.onCompleted: {
               const p = leaderPath.pointAtPercent(NGT.relativePosition(NGT.Track.station));
               x = trackShape.x + p.x - width;
               y = p.y + 3;
            }

            DLRLabel {
               anchors {
                  verticalCenter: parent.verticalCenter
                  right: parent.left
                  rightMargin: 2
               }

               text: "Station"
               font.pointSize: 7
            }
         }

         Rectangle {
            id: minimapLeader
            height: 7
            radius: 3
            width: 12
            color: DLRTheme.teaserYellow
            border.color: "black"

            DLRLabel {
               anchors {
                  horizontalCenter: parent.horizontalCenter
                  bottom: parent.top
               }

               text: "1"
               font.pointSize: 8
            }
         }

         Rectangle {
            id: minimapFollower
            height: 7
            radius: 3
            width: 12
            color: DLRTheme.teaserYellow
            border.color: "black"

            DLRLabel {
               anchors {
                  horizontalCenter: parent.horizontalCenter
                  top: parent.bottom
               }

               text: "2"
               font.pointSize: 8
            }
         }

         Signal {
            id: mmEntrySignal
            scale: 0.4

            Component.onCompleted: {
               const p = leaderPath.pointAtPercent(NGT.relativePosition(NGT.Track.entrySwitch));
               x = trackShape.x + p.x - width * 0.5;
               y = p.y - height / 1.3;
            }
         }

         Signal {
            id: mmExitSignal
            scale: 0.4

            Component.onCompleted: {
               const p = leaderPath.pointAtPercent(NGT.relativePosition(NGT.Track.exitSwitch));
               x = trackShape.x + p.x - width * 0.5;
               y = p.y - height / 1.3;
            }
         }
      }
   }

   function update() {
      trackTrainB.trackPositionMeter = NGT.simulation.leader.position;
      trackTrainA.trackPositionMeter = NGT.simulation.follower.position;

      entrySignal.open = !NGT.simulation.entrySwitchBlocked;
      exitSignal.open = !NGT.simulation.exitSwitchBlocked;

      mmEntrySignal.open = !NGT.simulation.entrySwitchBlocked;
      mmExitSignal.open = !NGT.simulation.exitSwitchBlocked;

      leaderBrakePoints.brakeDistanceMeter = NGT.simulation.brakeDistance(NGT.simulation.leader.speed);
      followerBrakePoints.brakeDistanceMeter = NGT.simulation.brakeDistance(NGT.simulation.follower.speed);

      {
         const leaderPos = NGT.relativePosition(NGT.simulation.leader.position);
         const p = leaderPath.pointAtPercent(leaderPos);
         minimapLeader.x = trackShape.x + p.x - minimapLeader.width;
         minimapLeader.y = p.y - minimapLeader.height / 2;
      }
      {
         const followerPos = NGT.relativePosition(NGT.simulation.follower.position);
         const p = followerPath.pointAtPercent(followerPos);
         minimapFollower.x = trackShape.x + p.x - minimapFollower.width;
         minimapFollower.y = p.y - minimapFollower.height / 2;
      }
   }

   function restart() {
      NGT.createSimulation(NGT.Scenario[radioGroup.checkedButton.text.toUpperCase()], speed.current, switchTime.current);
      simLoop.wait = 0;
      simLoop.tick = 0;
   }

   Timer {
      id: simLoop
      interval: 1
      running: true
      repeat: true

      property real wait: 0
      property var tick: 0

      onTriggered: {
         if (NGT.simulation.finished) {
            if (wait < 3000) {
               wait += interval;
            } else {
               view.restart();
            }
         } else {
            tick += 1;
            NGT.simulation.simulate((interval / 1000) * simulationSpeed.current);
            if (tick % 12 === 0) {
               update();
            }
         }
      }
   }

   DLRButton {
      anchors {
         bottom: parent.bottom
         right: parent.right
         rightMargin: horizontalMargin
         bottomMargin: 1.5 * verticalMargin + overviewButton.height * widgetScale
      }

      scale: widgetScale
      transformOrigin: Item.BottomRight

      text: "Show Info"
      onClicked: scenarioDescription.open()
   }

   ScenarioDescription {
      id: scenarioDescription
      z: 3000

      title: "Switches"
      descriptionURL: ":/gui/switches/description.html"

      onOpened: simulationSpeed.current = 0
      onClosed: simulationSpeed.current = 1
   }

   ReturnToOverviewButton {
      id: overviewButton
      z: 3001

      anchors {
         bottom: parent.bottom
         right: parent.right
         rightMargin: horizontalMargin
         bottomMargin: verticalMargin
      }
   }
}
