const Scenario = {
   "TRADITIONAL": 1,
   "SMART": 2,
   "PASSIVE": 3,
   "FAILSAFE": 4
}

/** @type m */
const TRAIN_LENGTH = 84;

/** @type m/s^2 */
const ACCELERATION = 1;

/** @type s */
const DELAY = 0.1;

/** @type m */
const DEFAULT_SWITCH_BLOCK_LENGTH = 5000;

/** @type s */
const MIN_STOP_TIME = 60;

/**
 * @param speed{m/s}
 * @param position{m}
 */
function Train(speed, position) {
   /** @type m/s */
   this.speed = speed;

   /** @type m */
   this.position = position;

   return this;
}

function SimulationParameters(scenario, initialSpeed, switchTime) {
   /** @type Scenario */
   this.currentScenario = scenario;

   /** @type s */
   this.switchTime = switchTime;

   /** @type m/s */
   this.initialSpeed = initialSpeed / 3.6;

   /** @type m/s */
   this.finalSpeed = this.initialSpeed;

   /** @type m */
   this.switchBlockLength = 0;
   if (this.currentScenario === Scenario.TRADITIONAL) {
      this.switchBlockLength = Math.max(this.finalSpeed * this.finalSpeed, this.initialSpeed * this.initialSpeed) / ACCELERATION;
   }

   /** @type s */
   this.entrySwitchTime = 0;
   if (this.currentScenario === Scenario.TRADITIONAL || this.currentScenario === Scenario.SMART) {
      this.entrySwitchTime = this.switchTime;
   }

   /** @type s */
   this.exitSwitchTime = 0;
   if (this.currentScenario === Scenario.TRADITIONAL ||
      this.currentScenario === Scenario.SMART ||
      this.currentScenario === Scenario.PASSIVE) {
      this.exitSwitchTime = this.switchTime;
   }

   /** @type m */
   this.initialBrakeDistance = (this.initialSpeed * DELAY) + ((this.initialSpeed * this.initialSpeed) / (2 * ACCELERATION));
   /** @type m */
   this.finalBrakeDistance = (this.finalSpeed * DELAY) + ((this.finalSpeed * this.finalSpeed) / (2 * ACCELERATION));

   /** @type m */
   this.initialRelativeDistance = Math.max(this.initialBrakeDistance + (this.exitSwitchTime * this.initialSpeed), this.switchBlockLength);
   /** @type m */
   this.finalRelativeDistance = Math.max(this.finalBrakeDistance + (this.entrySwitchTime * this.finalSpeed), this.switchBlockLength);

   /** @type s */
   this.initialRelativeTime = this.initialRelativeDistance / this.initialSpeed;
   /** @type s */
   this.finalRelativeTime = this.finalRelativeDistance / this.finalSpeed;

   /** @type s */
   this.leaderStopTime = MIN_STOP_TIME + (this.initialRelativeTime - this.finalRelativeTime);
   /** @type s */
   this.followerStopTime = MIN_STOP_TIME;

   /** @type s^-1 */
   this.maxFrequency = 1 / Math.max(this.initialRelativeTime, this.finalRelativeTime);
}

const Track = {
   /** @type m */
   start: 0,

   /** @type m */
   entrySwitch: 10716,

   /** @type m */
   station: 12768,

   /** @type m */
   exitSwitch: 12996,

   /** @type m */
   end: 23484
};

function relativePosition(p) {
   return p / Track.end;
}

/**
 * @param scenario{Scenario}
 * @param initialSpeed{km/h}
 * @param switchTime{s}
 */
function Simulation(scenario, initialSpeed, switchTime) {
   /** @type SimulationParameters */
   this.simParameters = new SimulationParameters(scenario, initialSpeed, switchTime);

   /** @type Train */
   this.leader = new Train(this.simParameters.initialSpeed, this.simParameters.initialRelativeDistance + TRAIN_LENGTH);
   /** @type Train */
   this.follower = new Train(this.simParameters.initialSpeed, 2 * TRAIN_LENGTH);

   /** @type m */
   this.startBrake = Track.station - this.simParameters.initialBrakeDistance;
   /** @type m */
   this.endAcceleration = Track.station + this.simParameters.initialBrakeDistance;

   /** @type boolean */
   this.entrySwitchBlocked = false;
   /** @type s */
   this.entrySwitchBlockTime = 0;
   /** @type string */
   this.entrySwitchPosition = "BOTTOM";

   /** @type boolean */
   this.exitSwitchBlocked = false;
   /** @type s */
   this.exitSwitchBlockTime = 0;
   /** @type string */
   this.exitSwitchPosition = "BOTTOM";

   /** @type boolean */
   this.leaderPause = false;
   /** @type s */
   this.leaderPauseTime = 0;
   /** @type boolean */
   this.leaderFinishedPause = false;

   /** @type boolean */
   this.followerPause = false;
   /** @type s */
   this.followerPauseTime = 0;
   /** @type boolean */
   this.followerFinishedPause = false;

   /** @type boolean */
   this.finished = false;

   /** @type s */
   this.elapsedTime = 0;

   /** @param deltaTime{s} */
   this.simulate = function (deltaTime) {
      this.elapsedTime += deltaTime;
      { // update leader
         if (!this.leaderPause && this.leader.position >= Track.station) {
            this.leaderPause = true;
         }

         if (this.leaderPause && !this.leaderFinishedPause) {
            if (this.leaderPauseTime > this.simParameters.leaderStopTime) {
               this.leaderFinishedPause = true;
            } else {
               this.leaderPauseTime += deltaTime;
            }
         }

         if (!this.leaderPause || this.leaderFinishedPause) {
            if (this.leader.position < this.startBrake || this.leader.position > this.endAcceleration) {
               // Do nothing
            } else if (this.leader.position >= Track.station) {
               this.leader.speed += ACCELERATION * deltaTime;
            } else if (this.leader.position > this.startBrake) {
               this.leader.speed -= ACCELERATION * deltaTime;

               if (this.leader.speed < 0) {
                  this.leader.speed = 0;
                  this.leader.position = Track.station;
               }
            }

            // update speed
            this.leader.position += this.leader.speed * deltaTime;
         }
      }

      { // update follower
         if (!this.followerPause && this.follower.position >= Track.station) {
            this.followerPause = true;
         }

         if (this.followerPause && !this.followerFinishedPause) {
            if (this.followerPauseTime > this.simParameters.followerStopTime) {
               this.followerFinishedPause = true;
            } else {
               this.followerPauseTime += deltaTime;
            }
         }

         if (!this.followerPause || this.followerFinishedPause) {
            if (this.follower.position < this.startBrake || this.follower.position > this.endAcceleration) {
               // Do nothing
            } else if (this.follower.position >= Track.station) {
               this.follower.speed += ACCELERATION * deltaTime;
            } else if (this.follower.position > this.startBrake) {
               this.follower.speed -= ACCELERATION * deltaTime;

               if (this.follower.speed < 0) {
                  this.follower.speed = 0;
                  this.follower.position = Track.station;
               }
            }

            // update speed
            this.follower.position += this.follower.speed * deltaTime;

         }
      }

      { // update entry switch
         if (!this.entrySwitchBlocked && this.entrySwitchPosition === "BOTTOM" && this.leader.position > Track.entrySwitch) {
            this.entrySwitchBlockTime = this.simParameters.entrySwitchTime;
            this.entrySwitchBlocked = true;
         }

         if (this.entrySwitchBlocked) {
            this.entrySwitchBlockTime -= deltaTime;
            if (this.entrySwitchBlockTime <= 0) {
               this.entrySwitchBlocked = false;
               this.entrySwitchPosition = "TOP";
            }
         }
      }

      { // update exit switch
         if (!this.exitSwitchBlocked && this.exitSwitchPosition === "BOTTOM" && this.leader.position > Track.exitSwitch) {
            this.exitSwitchBlockTime = this.simParameters.exitSwitchTime;
            this.exitSwitchBlocked = true;
         }

         if (this.exitSwitchBlocked) {
            this.exitSwitchBlockTime -= deltaTime;
            if (this.exitSwitchBlockTime <= 0) {
               this.exitSwitchBlocked = false;
               this.exitSwitchPosition = "TOP";
            }
         }
      }

      this.finished = this.leader.position > Track.end;
   }

   /**
    * @param speed{m/s}
    * @return {m}
    */
   this.brakeDistance = function (speed) {
      return speed * DELAY + (speed * speed) / (2 * ACCELERATION);
   }
}

let simulation = new Simulation(Scenario.TRADITIONAL, 200, 10);

/**
 * @param scenario{Scenario}
 * @param initialSpeed{km/h}
 * @param switchTime{s}
 */
function createSimulation(scenario, initialSpeed, switchTime) {
   simulation = new Simulation(scenario, initialSpeed, switchTime);
}

/**
 * @param scenario{Scenario}
 * @param initialSpeed{km/h}
 * @param switchTime{s}
 */
function getFrequency(scenario, initialSpeed, switchTime) {
   return new SimulationParameters(scenario, initialSpeed, switchTime).maxFrequency;
}