#include <QtGui/QtGui>
#include <QtQml/QtQml>
#include <QtQuick/QtQuick>
#include <QtQuickControls2/QtQuickControls2>
#include <QtWidgets/QApplication>

#include <spdlog/spdlog.h>

#include <cstdint>
#include <exception>

#include <ngt/utils/AppInfo.hpp>
#include <ngt/utils/FileReader.hpp>
#include <ngt/utils/MacroWrappers.hpp>
#include <ngt/utils/StringUtil.hpp>
#include <ngt/views/overview/Overview.hpp>
#include <ngt/views/speed-limits/SpeedLimits.hpp>

using namespace ngt; // NOLINT

#ifdef _WIN64
extern "C" {
// This tells Windows to use the dedicated NVIDIA GPU over Intel integrated graphics.
__declspec(dllexport) uint32_t NvOptimusEnablement = 0x00000001; // NOLINT

// This tells Windows to use the dedicated AMD GPU over Intel integrated graphics.
__declspec(dllexport) int AmdPowerXpressRequestHighPerformance = 1; // NOLINT
}
#endif

auto main(int argc, char* argv[]) -> int {
   std::vector<char*> args{argv, argv + argc};

   args.push_back(const_cast<char*>("-platform"));
   if constexpr (utils::HostPlatform == utils::Platform::Windows) {
      args.push_back(const_cast<char*>("windows:dpiawareness=0"));
   } else if constexpr (utils::HostPlatform == utils::Platform::Web) {
      args.push_back(const_cast<char*>("wasm:dpiawareness=0"));
   }

   try {
      spdlog::set_pattern("%^[%L] %$%v");

      if constexpr (utils::DEBUG_ENABLED) {
         spdlog::set_level(spdlog::level::debug);
      }

      qInstallMessageHandler([](QtMsgType type, QMessageLogContext const& context, QString const& message) {
         std::string file;
         if (context.file != nullptr) {
            file     = context.file;
            size_t i = file.find(std::string(context.category) == "qml" ? "resources/" : "src/");
            if (i != std::string::npos) {
               file = file.substr(i);
            }
         }

         std::string msg = message.toStdString();
         replaceString(msg, "qrc:", "resources");

         const char* function = context.function != nullptr ? context.function : "";
         switch (type) {
            case QtDebugMsg:
               spdlog::debug("{} ({}:{}, {})", msg, file, context.line, function);
               break;
            case QtInfoMsg:
               spdlog::info(msg);
               break;
            case QtWarningMsg:
               spdlog::warn(msg);
               break;
            case QtCriticalMsg:
               spdlog::error(msg);
               break;
            case QtFatalMsg:
               spdlog::critical("{} ({}:{}, {})", msg, file, context.line, function);
               break;
            default:
               spdlog::info(msg);
         }
      });

      qputenv("QML_XHR_ALLOW_FILE_READ", QByteArray("1"));
      qputenv("QT_SCALE_FACTOR", QByteArray("1"));

      spdlog::info("Starting Demonstrator!");
      int size = static_cast<int>(args.size());
      QApplication::setAttribute(Qt::AA_DisableHighDpiScaling);
      QApplication app(size, args.data());

      spdlog::info("Registering Fonts.");
      QFontDatabase::addApplicationFont(":/fonts/ARIAL.TTF");
      QFontDatabase::addApplicationFont(":/fonts/ARIALBD.TTF");
      QFontDatabase::addApplicationFont(":/fonts/ARIALBI.TTF");
      QFontDatabase::addApplicationFont(":/fonts/ARIALI.TTF");
      QApplication::setFont(QFont("Arial", 14));

      spdlog::info("Setting window icon.");
      QGuiApplication::setWindowIcon(QIcon(":/images/DLRSignet.png"));

      spdlog::info("Registering types.");
      qmlRegisterType<overview::Overview>("Overview", 1, 0, "OverviewModel");
      qmlRegisterType<speedlimits::SpeedLimits>("SpeedLimits", 1, 0, "SpeedLimitsModel");
      qRegisterMetaType<speedlimits::TrackSegmentListViewModel*>("TrackSegmentListViewModel*");
      qRegisterMetaType<speedlimits::TrainViewModel*>("TrainViewModel*");
      qRegisterMetaType<speedlimits::TrainListViewModel*>("TrainListViewModel*");
      qmlRegisterType<io::FileReader>("IO", 1, 0, "FileReader");
      qmlRegisterType<utils::AppInfo>("App", 1, 0, "AppInfo");

      spdlog::info("Setting format.");
      QSurfaceFormat format;
      format.setSamples(8);
      QSurfaceFormat::setDefaultFormat(format);

      spdlog::info("Starting Engine.");
      QQmlApplicationEngine appEngine{};

      spdlog::info("Loading main qml file.");
      appEngine.load(QUrl("qrc:/gui/main.qml"));
      if (appEngine.rootObjects().isEmpty()) {
         spdlog::critical("Couldn't load the qml resource! Terminating.");
         return -1;
      }

      spdlog::info("Trains departing!");
      return QGuiApplication::exec();
   } catch (std::exception const& e) {
      try {
         spdlog::critical(e.what());
      } catch (...) {
         std::cerr << "There seems to be an issue with logging via 'spdlog'. Logging with 'iostream' instead.\n"
                   << "The original error is:\n"
                   << e.what() << std::endl;
      }
      return -1;
   }
}
