#include "JsonLoader.hpp"
#include "JsonModels.hpp"

#include <chrono>
#include <ngt/utils/io.hpp>
#include <ngt/views/overview/model/Station.hpp>
#include <nlohmann/json.hpp>
#include <set>

namespace ngt::overview {

auto parseNodes(nlohmann::json const& j) -> std::pair<StationList, NodeMap> {
   auto nodes = j.get<std::map<std::string, json::Node>>();

   StationList stations{};
   NodeMap     trackNodes{};

   for (auto const& [_, node] : nodes) {
      Coordinates coordinates{node.coordinates.second, node.coordinates.first};
      trackNodes.try_emplace(node.name, node.name, coordinates);

      if (node.isStation) {
         stations.emplace_back(node.name, node.shortName, coordinates, node.displayName, node.left);
      }
   }

   return {stations, trackNodes};
}

auto parseEdges(nlohmann::json const& j, NodeMap const& nodes) -> TrackMap {
   auto edges = j.get<std::map<std::string, json::Edge>>();

   TrackMap trackMap{};

   for (auto const& [key, edge] : edges) {
      auto const& start = nodes.at(edge.start);
      auto const& end   = nodes.at(edge.end);
      trackMap.try_emplace(edge.id, edge.id, start, end, edge.length);
   }

   return trackMap;
}

auto loadTrainNetwork(std::string_view file) -> std::tuple<StationList, NodeMap, TrackMap> {
   nlohmann::json j = io::loadJson(file);

   auto [stations, nodes] = parseNodes(j.at("nodes"));
   TrackMap tracks        = parseEdges(j.at("edges"), nodes);

   return {stations, nodes, tracks};
}

auto loadTrainData(std::string_view file)
   -> std::pair<std::unordered_map<std::string, TrainData>, std::chrono::milliseconds> {
   using namespace std::chrono_literals;

   nlohmann::json j         = io::loadJson(file);
   auto           positions = j.get<std::map<std::string, json::TrainPositions>>();

   std::unordered_map<std::string, TrainData> trainData{};

   std::chrono::milliseconds maxTime = 0ms;

   for (auto const& [id, pos] : positions) {
      std::map<std::chrono::milliseconds, TimeStepData> timeSteps{};

      for (auto const& timeStep : pos.timeStamps) {
         double const              toMillis = 1000;
         std::chrono::milliseconds t{std::llround(timeStep.time * toMillis)};
         maxTime = std::max(t, maxTime);
         timeSteps.try_emplace(t, t, timeStep.edge, timeStep.edgePosition);
      }

      trainData.try_emplace(id, id, timeSteps);
   }

   return {trainData, maxTime};
}

auto loadTrains(std::string_view file) -> std::vector<Train> {
   nlohmann::json j          = io::loadJson(file);
   auto           jsonTrains = j.get<std::map<std::string, json::Train>>();

   std::vector<Train> trains;
   trains.reserve(jsonTrains.size());

   for (auto const& [_, train] : jsonTrains) {
      trains.emplace_back(train.id, train.waggons);
   }

   return trains;
}

} // namespace ngt::overview