#include "Simulator.hpp"
#include "JsonLoader.hpp"

namespace ngt::overview {

auto getNetworkData() -> std::tuple<StationList, NodeMap, TrackMap>& {
   static std::tuple<StationList, NodeMap, TrackMap> NETWORK = loadTrainNetwork(":/data/overview/network.json");
   return NETWORK;
}

auto getTrainData() -> std::vector<Train>& {
   static std::vector<Train> TRAINS = loadTrains(":/data/overview/trains.json");
   return TRAINS;
}

using TrainPositons = std::pair<std::unordered_map<std::string, TrainData>, std::chrono::milliseconds>;

auto getPositionData(uint8_t scenario) -> TrainPositons& {
   static std::array<TrainPositons, 3> DATA{loadTrainData(":/data/overview/log_pos_base.json"),
                                            loadTrainData(":/data/overview/log_pos_single.json"),
                                            loadTrainData(":/data/overview/log_pos_double.json")};
   return DATA.at(scenario);
}

Simulator::Simulator(uint8_t scenario)
   : scenario(scenario)
   , maxTime() {
   stations      = &std::get<StationList>(getNetworkData());
   trackNodes    = &std::get<NodeMap>(getNetworkData());
   trackSegments = &std::get<TrackMap>(getNetworkData());

   maxTime   = getPositionData(scenario).second;
   trainData = &getPositionData(scenario).first;

   trains = &getTrainData();
}

Simulator::Simulator(const Simulator& other)
   : Simulator(other.scenario) {}

Simulator::Simulator(Simulator&& other) noexcept
   : scenario(other.scenario)
   , maxTime(other.maxTime) {
   stations      = other.stations;
   trackNodes    = other.trackNodes;
   trackSegments = other.trackSegments;
   trainData     = other.trainData;
   trains        = other.trains;
}

auto Simulator::operator=(const Simulator& other) -> Simulator& {
   if (this != &other) {
      *this = Simulator(other);
   }

   return *this;
}

auto Simulator::operator=(Simulator&& other) noexcept -> Simulator& {
   if (this == &other) {
      return *this;
   }

   scenario = other.scenario;
   maxTime  = other.maxTime;

   stations      = other.stations;
   trackNodes    = other.trackNodes;
   trackSegments = other.trackSegments;
   trainData     = other.trainData;
   trains        = other.trains;

   return *this;
}

void Simulator::startSimulation() {
   using namespace std::chrono_literals;
   simulationTime = 0ms;
}

auto Simulator::coordinatesFromTrackPosition(std::string_view trackId, double position) const noexcept -> Coordinates {
   auto const& track = trackSegments->at(trackId.data());
   auto const& start = trackNodes->at(track.start.id);
   auto const& end   = trackNodes->at(track.end.id);

   double positionNorm = position / track.length;
   auto   pos          = glm::mix(start.coordinates.toVec2(), end.coordinates.toVec2(), positionNorm);
   return Coordinates{pos};
}

void Simulator::nextTick(std::chrono::milliseconds deltaT) {
   using namespace std::chrono_literals;

   if (simulationTime > maxTime) {
      simulationTime = 0ms;
   }

   simulate(simulationTime);
   double const toSeconds = 1.0 / 1000.0;
   emit         simulationUpdate(static_cast<int32_t>(simulationTime.count() * toSeconds));

   simulationTime += deltaT * std::llround(speed);
}

void Simulator::simulate(std::chrono::milliseconds t) {
   for (auto&& train : *trains) {
      if (trainData->find(train.id) == trainData->end()) {
         if (train.coordinates.latitude != 0.0 || train.coordinates.longitude != 0.0) {
            train.coordinates = Coordinates(0.0, 0.0);
            for (auto&& wagon : train.wagons) {
               wagon.coordinates = Coordinates(0.0, 0.0);
            }
         }
         continue;
      }

      auto const& data = trainData->at(train.id);

      auto const& timeStamp = data.getTimeStep(t);
      train.coordinates     = coordinatesFromTrackPosition(timeStamp.trackId, timeStamp.position);

      for (size_t i = 0; i < train.wagons.size(); ++i) {
         double positionInTrain  = (WAGON_LENGTH / 2) + i * (WAGON_GAP + WAGON_LENGTH);
         double absolutePosition = timeStamp.position - positionInTrain;

         std::string trackId = timeStamp.trackId;
         if (absolutePosition < 0) {
            auto const& prev = data.findPreviousEdge(t);
            if (prev) {
               trackId          = prev->trackId;
               absolutePosition = trackSegments->at(trackId).length + absolutePosition;
            }
         } else if (absolutePosition > trackSegments->at(trackId).length) {
            TODO("THIS SHOULDN'T HAPPEN!?");
         }

         train.wagons[i].coordinates = coordinatesFromTrackPosition(trackId, absolutePosition);
      }
   }
}

void Simulator::stopSimulation() {}

void Simulator::setSpeed(double newSpeed) {
   speed = newSpeed;
}

} // namespace ngt::overview
