#include "SpeedLimits.hpp"
#include <spdlog/spdlog.h>

namespace ngt::speedlimits {

SpeedLimits::SpeedLimits() {
   mcts              = TrainViewModel(simulator.getMCTS());
   vcts              = TrainListViewModel(simulator.getVCTS());
   trackSegments     = TrackSegmentListViewModel(simulator.getTrack());
   nextTrackSegments = TrackSegmentListViewModel(simulator.getNextRunTrack());

   connect(this, &SpeedLimits::setSpeed, &simulator, &Simulator::setSpeed);
   connect(this, &SpeedLimits::setNextRunTrains, &simulator, &Simulator::setNextRunNumberOfTrains);
   connect(this, &SpeedLimits::setNextRunTrainLength, &simulator, &Simulator::setNextRunTrainLength);
   connect(this, &SpeedLimits::restart, &simulator, &Simulator::restart);
   connect(&simulator, &Simulator::simulationUpdate, this, &SpeedLimits::simulationUpdate);

   startSimulation();
}

SpeedLimits::~SpeedLimits() {
   stopSimulation();
}

void SpeedLimits::startSimulation() noexcept {
   simulator.startSimulation();
}

void SpeedLimits::stopSimulation() noexcept {
   simulator.stopSimulation();
}

void SpeedLimits::simulationUpdate(int t) noexcept {
   if (t < lastTime) {
      emit restart();
      emit trackSegments.layoutChanged();
   }

   if (t != lastTime) {
      pSimulationTime = t;
      lastTime        = t;
   }

   vcts.changeEvent();
   emit mcts.change();
   emit update();
}

void SpeedLimits::setSimulationSpeed(double newSpeed) noexcept {
   if (pSimulationSpeed != newSpeed) {
      pSimulationSpeed = newSpeed;
      emit setSpeed(newSpeed);
   }
}

auto SpeedLimits::getMCTS() noexcept -> TrainViewModel* {
   return &mcts;
}

auto SpeedLimits::getVCTS() noexcept -> TrainListViewModel* {
   return &vcts;
}

auto SpeedLimits::getTrack() noexcept -> TrackSegmentListViewModel* {
   return &trackSegments;
}

auto SpeedLimits::getNextTrack() noexcept -> TrackSegmentListViewModel* {
   return &nextTrackSegments;
}

auto SpeedLimits::getBounds() noexcept -> QPointF {
   auto [min, max] = simulator.getBounds();
   return {min.to<double>(), max.to<double>()};
}

void SpeedLimits::simulate(int deltaT) {
   simulator.nextTick(std::chrono::milliseconds(deltaT));
}

} // namespace ngt::speedlimits