#pragma once

#include <QtCharts/QtCharts>
#include <QtCore/QObject>
#include <QtCore/QRunnable>
#include <QtQml/QtQml>

#include <ngt/views/speed-limits/control/Simulator.hpp>
#include <ngt/views/speed-limits/view-model/TrackSegmentListViewModel.hpp>
#include <ngt/views/speed-limits/view-model/TrainListViewModel.hpp>
#include <ngt/views/speed-limits/view-model/TrainViewModel.hpp>

namespace ngt::speedlimits {

class SpeedLimits : public QObject {
   Q_OBJECT; // NOLINT

   Q_PROPERTY(int simulationTime READ simulationTime NOTIFY update);
   Q_PROPERTY(double simulationSpeed READ simulationSpeed WRITE setSimulationSpeed);

   Q_PROPERTY(double trainLength READ trainLength WRITE setTrainLength);
   Q_PROPERTY(int numTrains READ numTrains WRITE setNumTrains);

   Q_PROPERTY(TrackSegmentListViewModel* track READ getTrack NOTIFY restart);
   Q_PROPERTY(TrackSegmentListViewModel* nextTrack READ getNextTrack NOTIFY restart);

   Q_PROPERTY(TrainViewModel* mcts READ getMCTS NOTIFY update);
   Q_PROPERTY(TrainListViewModel* vcts READ getVCTS NOTIFY update);
   Q_PROPERTY(QPointF bounds READ getBounds NOTIFY update);

   Q_PROPERTY(std::vector<double> mctsSpeeds READ getMCTSSpeeds NOTIFY update);
   Q_PROPERTY(QList<QList<double>> vctsSpeeds READ getVCTSSpeeds NOTIFY update);

   Q_PROPERTY(double vctsTime READ getVCTSTime NOTIFY update);
   Q_PROPERTY(double mctsTime READ getMCTSTime NOTIFY update);

 public:
   SpeedLimits();

   SpeedLimits(SpeedLimits const& other) = delete;
   auto operator=(SpeedLimits const& other) -> SpeedLimits& = delete;

   SpeedLimits(SpeedLimits&& other) = delete;
   auto operator=(SpeedLimits&& other) -> SpeedLimits& = delete;

   ~SpeedLimits() override;

   void startSimulation() noexcept;
   void stopSimulation() noexcept;

   [[nodiscard]]
   auto simulationTime() const noexcept -> int { return pSimulationTime; }

   [[nodiscard]]
   auto simulationSpeed() const noexcept -> double { return pSimulationSpeed; }

   void setSimulationSpeed(double newSpeed) noexcept;

   [[nodiscard]]
   auto getMCTS() noexcept -> TrainViewModel*;

   [[nodiscard]]
   auto getVCTS() noexcept -> TrainListViewModel*;

   [[nodiscard]]
   auto getTrack() noexcept -> TrackSegmentListViewModel*;

   [[nodiscard]]
   auto getNextTrack() noexcept -> TrackSegmentListViewModel*;

   [[nodiscard]]
   auto getBounds() noexcept -> QPointF;

   [[nodiscard]]
   auto numTrains() const noexcept -> int { return simulator.getNextRunNumberOfTrains(); }
   void setNumTrains(int numTrains) noexcept { emit setNextRunTrains(static_cast<uint8_t>(numTrains)); }

   [[nodiscard]]
   auto trainLength() const noexcept -> double { return simulator.getNextRunTrainLength(); }
   void setTrainLength(double length) noexcept { emit setNextRunTrainLength(length); }

   [[nodiscard]]
   auto getMCTSSpeeds() noexcept -> std::vector<double>& { return *simulator.getMCTSSpeeds(); }

   [[nodiscard]]
   auto getVCTSSpeeds() noexcept -> QList<QList<double>>& { return *simulator.getVCTSSpeeds(); }

   [[nodiscard]]
   auto getVCTSTime() noexcept -> double { return simulator.getVCTSTime().value(); }

   [[nodiscard]]
   auto getMCTSTime() noexcept -> double { return simulator.getMCTSTime().value(); }

   Q_INVOKABLE void nextRun() { emit restart(); }

   Q_INVOKABLE void simulate(int deltaT);

 signals:
   void update();
   void restart();

 private:
   TrainViewModel            mcts;
   TrainListViewModel        vcts;
   TrackSegmentListViewModel trackSegments;
   TrackSegmentListViewModel nextTrackSegments;

   Simulator simulator{};

   int pSimulationTime = 0;
   int lastTime        = 0;

   double pSimulationSpeed = 10.0;

 public slots:
   void simulationUpdate(int t) noexcept;

 signals:
   void setSpeed(double _t1);
   void setNextRunTrains(uint8_t num);
   void setNextRunTrainLength(double length);
};

} // namespace ngt::speedlimits
