#include "Simulator.hpp"

#include <ngt/utils/si.hpp>

#include <spdlog/spdlog.h>

namespace ngt::speedlimits {

Simulator::Simulator()
   : track(std::make_unique<Track>(loadFromJson(":/data/speed-limits/speed-limits.json").track))
   , nextRunTrack(std::make_unique<Track>(loadFromJson(":/data/speed-limits/speed-limits.json").track))
   , trainProps(loadFromJson(":/data/speed-limits/speed-limits.json").trainProperties)
   , nextRunTrainLength(trainProps.length)
   , nextRunNumberOfTrains(trainProps.wagons)
   , mcts(std::make_unique<Train>(0_km, 0_kph, trainProps.length * trainProps.wagons))
   , vcts(std::make_unique<std::vector<Train>>()) {

   for (uint8_t i = 0U; i < trainProps.wagons; ++i) {
      vcts->emplace_back(-i * (120_m + trainProps.length), 0_kph, trainProps.length);
      vctsSpeeds += QList<double>();
   }
}

void Simulator::reset() {
   auto const startSpeed = track->getSectionAt(0_km).speedLimit;

   vctsFinished            = false;
   mechanicalTrainFinished = false;

   trainProps.length = nextRunTrainLength;
   trainProps.wagons = nextRunNumberOfTrains;
   *track            = *nextRunTrack;

   mcts->position = 0_km;
   mcts->speed    = startSpeed;
   mcts->length   = trainProps.length * trainProps.wagons;

   vcts->clear();
   vctsSpeeds.clear();
   for (uint8_t i = 0U; i < trainProps.wagons; ++i) {
      vcts->emplace_back(-i * (120_m + trainProps.length), startSpeed, trainProps.length);
      vctsSpeeds += QList<double>();
   }

   mctsSpeeds.clear();
}

void Simulator::restart() {
   using namespace std::chrono_literals;
   finished       = false;
   simulationTime = 0ms;
   lastFrame      = 0ms;
   lastSecond     = 0s;
   reset();
   calculateTimes();
}

auto getSectionTime(si::meter_t length, si::meters_per_second_t vStart, si::meters_per_second_t vEnd,
                    si::meters_per_second_t vMax) -> si::second_t {
   si::second_t tAccel = si::math::abs(vMax - vStart) / 1_mps_sq;
   si::meter_t  lAccel = tAccel * (vMax + vStart) / 2;

   si::second_t tDecel = si::math::abs(vMax - vEnd) / 1_mps_sq;
   si::meter_t  lDecel = tDecel * (vMax + vEnd) / 2;

   si::meter_t  lConst = length - (lAccel + lDecel);
   si::second_t tConst = lConst / vMax;

   return tAccel + tConst + tDecel;
}

auto Simulator::calculateTime(Train const& train, size_t positionInSet) -> si::second_t {
   si::second_t completeTime = 0_s;
   si::meter_t  startOffset  = train.position;

   auto lastSpeed = track->getSectionAt(0_km).speedLimit;
   for (auto const& [km, section] : track->getSections()) {
      auto nextSpeed = track->getNextSection(km).second.speedLimit;
      auto length    = section.length - startOffset;
      if (lastSpeed < section.speedLimit) {
         length -= train.length;
      }

      if (nextSpeed > section.speedLimit) {
         length += train.length;
      }

      completeTime += getSectionTime(length, si::math::min(lastSpeed, section.speedLimit),
                                     si::math::min(nextSpeed, section.speedLimit), section.speedLimit);
      lastSpeed   = section.speedLimit;
      startOffset = 0_m;
   }

   completeTime += getSectionTime(maxPossibleLength - (positionInSet * train.length), lastSpeed, 0_mps, lastSpeed);
   return completeTime;
}

void Simulator::calculateTimes() {
   mctsTime = calculateTime(*mcts, 0);
   vctsTime = calculateTime(vcts->back(), vcts->size() - 1);
}

void Simulator::startSimulation() {
   using namespace std::chrono_literals;

   mechanicalTrainFinished = false;
   vctsFinished            = false;
   finished                = false;
   simulationTime          = 0ms;

   auto const startSpeed = track->getSectionAt(0_km).speedLimit;

   mcts->speed = startSpeed;
   for (auto&& train : *vcts) {
      train.speed = startSpeed;
   }

   calculateTimes();
}

auto slowDownDistance(si::meters_per_second_t startSpeed, si::meters_per_second_t endSpeed,
                      si::meters_per_second_squared_t breakForce) -> si::meter_t {
   return (startSpeed * startSpeed - endSpeed * endSpeed) / (2.0 * -breakForce);
}

void Simulator::simulateTrain(Train& train, si::second_t elapsed) {
   TrackSection const& sectionFront = track->getSectionAt(train.position);
   TrackSection const& sectionBack  = track->getSectionAt(train.position - train.length);

   si::meters_per_second_t currentSpeedLimit = std::min(sectionBack.speedLimit, sectionFront.speedLimit);

   auto const [nextSectionStart, nextSection] = track->getNextSection(train.position);
   si::meter_t slowDownDist = slowDownDistance(train.speed, nextSection.speedLimit, trainProps.deceleration);

   if (nextSection.speedLimit < train.speed and slowDownDist > nextSectionStart - train.position) {
      currentSpeedLimit = nextSection.speedLimit;
   }

   auto   speedDiff = std::abs((currentSpeedLimit - train.speed).to<double>());
   double smooth    = std::min(speedDiff, 1.0);

   // We have to break if the train is too fast or a slower section is coming up.
   if (train.speed > currentSpeedLimit) {
      train.speed += trainProps.deceleration * smooth * elapsed;
   } else if (train.speed < currentSpeedLimit) {
      train.speed += trainProps.acceleration * smooth * elapsed;
   }

   train.position += train.speed * elapsed;
}

void Simulator::simulateArrival(Train& train, si::kilometer_t stoppingPoint, si::second_t elapsed) {
   auto currentSpeedLimit = track->getSections().rbegin()->second.speedLimit;

   // This is a small hack to avoid simulation delay. We force to break a little earlier.
   si::meter_t slowDownDist = slowDownDistance(train.speed, -2_kph, trainProps.deceleration);
   if (slowDownDist > stoppingPoint - train.position) {
      currentSpeedLimit = 0_kph;
   }

   // We have to break if the train is too fast or a slower section is coming up.
   if (train.speed > currentSpeedLimit) {
      train.speed += trainProps.deceleration * elapsed;
   } else if (train.speed < currentSpeedLimit) {
      train.speed += trainProps.acceleration * elapsed;
   }

   if (train.speed < 0_kph) {
      train.speed = 0_kph;
   }

   train.position += train.speed * elapsed;
}

void Simulator::nextTick(std::chrono::milliseconds deltaT) {
   using namespace std::chrono_literals;

   // When both trains finished wait 5 more seconds.
   if (mechanicalTrainFinished and vctsFinished and not finished) {
      finished  = true;
      resetTime = std::chrono::high_resolution_clock::now() + 3s;
   }

   if (not finished) {
      simulate(simulationTime);
      emit simulationUpdate(static_cast<int32_t>(simulationTime.count() / 1000));

      simulationTime += deltaT * std::llround(speed);
   }

   // When the 5 seconds passed reset the clock and start over.
   if (finished and std::chrono::high_resolution_clock::now() > resetTime) {
      restart();
   }
}

void Simulator::simulate(std::chrono::milliseconds t) {
   auto elapsed = si::second_t(t - lastFrame);

   bool newSecond = false;
   auto tSeconds  = std::chrono::duration_cast<std::chrono::seconds>(t);
   if (tSeconds > lastSecond) {
      lastSecond = tSeconds;
      newSecond  = true;
   }

   if (mcts->position < track->length()) {
      simulateTrain(*mcts, elapsed);
   } else if (not mcts->stopped()) {
      simulateArrival(*mcts, track->length() + maxPossibleLength, elapsed);
   } else {
      mechanicalTrainFinished = true;
   }

   if (newSecond) {
      mctsSpeeds.emplace_back(mcts->speed.to<double>());
   }

   bool allStopped = true;
   for (size_t i = 0; i < vcts->size(); ++i) {
      auto&& vctsTrain = vcts->at(i);
      if (vctsTrain.position < track->length()) {
         simulateTrain(vctsTrain, elapsed);
         allStopped = false;
      } else if (not vctsTrain.stopped()) {
         simulateArrival(vctsTrain, track->length() + maxPossibleLength - i * vctsTrain.length, elapsed);
         allStopped = false;
      }

      if (newSecond) {
         vctsSpeeds[static_cast<int>(i)] += vcts->at(i).speed.to<double>();
      }
   }

   if (allStopped) {
      vctsFinished = true;
   }

   lastFrame = t;
}

void Simulator::stopSimulation() {}

void Simulator::setSpeed(double newSpeed) {
   speed = newSpeed;
}

auto Simulator::getBounds() const noexcept -> std::pair<si::kilometer_t, si::kilometer_t> {
   si::kilometer_t min = std::min(mcts->position - mcts->length, vcts->back().position - vcts->back().length);
   si::kilometer_t max = std::max(mcts->position, vcts->front().position);
   return {min, max};
}

} // namespace ngt::speedlimits
