#pragma once

#include "JsonLoader.hpp"

#include <QtCore/QObject>

namespace ngt::speedlimits {
class Simulator : public QObject {
   Q_OBJECT; // NOLINT
 public:
   Simulator();

   void startSimulation();
   void nextTick(std::chrono::milliseconds deltaT);
   void simulate(std::chrono::milliseconds t);
   void stopSimulation();

   [[nodiscard]]
   auto getVCTS() const noexcept -> std::vector<Train> const* { return vcts.get(); }

   [[nodiscard]]
   auto getMCTS() const noexcept -> Train const* { return mcts.get(); }

   [[nodiscard]]
   auto getTrack() const noexcept -> Track* { return track.get(); }

   [[nodiscard]]
   auto getNextRunTrack() const noexcept -> Track* { return nextRunTrack.get(); }

   [[nodiscard]]
   auto getBounds() const noexcept -> std::pair<si::kilometer_t, si::kilometer_t>;

   [[nodiscard]]
   auto getMCTSSpeeds() noexcept -> std::vector<double>* { return &mctsSpeeds; }

   [[nodiscard]]
   auto getVCTSSpeeds() noexcept -> QList<QList<double>>* { return &vctsSpeeds; }

   [[nodiscard]]
   auto getNextRunTrainLength() const noexcept -> double { return nextRunTrainLength.to<double>(); }

   [[nodiscard]]
   auto getNextRunNumberOfTrains() const noexcept -> uint8_t { return nextRunNumberOfTrains; }

   [[nodiscard]]
   auto getVCTSTime() const noexcept -> si::second_t { return vctsTime; };

   [[nodiscard]]
   auto getMCTSTime() const noexcept -> si::second_t { return mctsTime; };

 private:
   TrainProperties trainProps;
   si::meter_t     nextRunTrainLength;
   uint8_t         nextRunNumberOfTrains;

   si::kilometer_t maxPossibleLength = 6 * 300_m;

   std::unique_ptr<Track> track;
   std::unique_ptr<Track> nextRunTrack;

   std::unique_ptr<std::vector<Train>> vcts;
   std::unique_ptr<Train>              mcts;

   bool mechanicalTrainFinished = false;
   bool vctsFinished            = false;

   std::chrono::milliseconds                                   simulationTime{};
   std::chrono::time_point<std::chrono::high_resolution_clock> resetTime;

   std::chrono::seconds      lastSecond{0};
   std::chrono::milliseconds lastFrame{0};
   double                    speed{10};
   bool                      finished{false};

   si::second_t mctsTime = 0_s;
   si::second_t vctsTime = 0_s;

   std::vector<double>  mctsSpeeds{};
   QList<QList<double>> vctsSpeeds{};

   void reset();
   auto calculateTime(const Train& train, size_t positionInSet) -> si::second_t;
   void calculateTimes();
   void simulateTrain(Train& train, si::second_t elapsed);
   void simulateArrival(Train& train, si::kilometer_t stoppingPoint, units::time::second_t elapsed);

 public slots:
   void setSpeed(double newSpeed);
   void setNextRunNumberOfTrains(uint8_t num) { nextRunNumberOfTrains = num; }
   void setNextRunTrainLength(double length) { nextRunTrainLength = si::meter_t(length); }
   void restart();

 signals:
   void simulationUpdate(int _t1);
};
}; // namespace ngt::speedlimits
